# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from unittest.mock import patch

from swh.storage.metrics import (
    send_metric, OPERATIONS_METRIC, OPERATIONS_UNIT_METRIC
)


def test_send_metric_unknown_unit():
    r = send_metric('content', count=10, method_name='content_add')
    assert r is False
    r = send_metric('sthg:add:bytes:extra', count=10, method_name='sthg_add')
    assert r is False


def test_send_metric_no_value():
    r = send_metric('content:add', count=0, method_name='content_add')
    assert r is False


@patch('swh.storage.metrics.statsd.increment')
def test_send_metric_no_unit(mock_statsd):
    r = send_metric('content:add', count=10, method_name='content_add')

    mock_statsd.assert_called_with(OPERATIONS_METRIC, 10, tags={
        'endpoint': 'content_add',
        'object_type': 'content',
        'operation': 'add',
    })

    assert r


@patch('swh.storage.metrics.statsd.increment')
def test_send_metric_unit(mock_statsd):
    unit_ = 'bytes'
    r = send_metric('c:add:%s' % unit_, count=100, method_name='c_add')

    expected_metric = OPERATIONS_UNIT_METRIC.format(unit=unit_)
    mock_statsd.assert_called_with(
        expected_metric, 100, tags={
            'endpoint': 'c_add',
            'object_type': 'c',
            'operation': 'add',
        })

    assert r



# Copyright (C) 2015-2017  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from swh.core.api import RPCClient

from ..exc import StorageAPIError
from ..storage import Storage


class RemoteStorage(RPCClient):
    """Proxy to a remote storage API"""
    api_exception = StorageAPIError
    backend_class = Storage

    def reset(self):
        return self.post('reset', {})

    def stat_counters(self):
        return self.get('stat/counters')

    def refresh_stat_counters(self):
        return self.get('stat/refresh')



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import Optional, Tuple


def _is_power_of_two(n: int) -> bool:
    return n > 0 and n & (n-1) == 0


def get_partition_bounds_bytes(
        i: int, n: int, nb_bytes: int) -> Tuple[bytes, Optional[bytes]]:
    r"""Splits the range [0; 2^(nb_bytes*8)) into n same-length intervals,
    and returns the boundaries of this interval (both inclusive); or None
    as upper bound, if this is the last partition

    n must be a power of 2.

    >>> get_partition_bounds_bytes(0, 16, 2) == (b'\x00\x00', b'\x10\x00')
    True
    >>> get_partition_bounds_bytes(1, 16, 2) == (b'\x10\x00', b'\x20\x00')
    True
    >>> get_partition_bounds_bytes(14, 16, 2) == (b'\xe0\x00', b'\xf0\x00')
    True
    >>> get_partition_bounds_bytes(15, 16, 2) == (b'\xf0\x00', None)
    True
    """
    if not _is_power_of_two(n):
        raise ValueError('number of partitions must be a power of two')
    if not 0 <= i < n:
        raise ValueError(
            'partition index must be between 0 and the number of partitions.')

    space_size = 1 << (nb_bytes*8)
    partition_size = space_size//n

    start = (partition_size*i).to_bytes(nb_bytes, 'big')
    end = None if i == n-1 \
        else (partition_size*(i+1)).to_bytes(nb_bytes, 'big')
    return (start, end)



# Copyright (C) 2015-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os
import logging

from swh.core import config
from swh.storage import get_storage as get_swhstorage
from swh.core.api import (RPCServerApp,
                          error_handler,
                          encode_data_server as encode_data)

from ..storage import Storage
from ..metrics import timed


def get_storage():
    global storage
    if not storage:
        storage = get_swhstorage(**app.config['storage'])

    return storage


app = RPCServerApp(__name__,
                   backend_class=Storage,
                   backend_factory=get_storage)
storage = None


@app.errorhandler(Exception)
def my_error_handler(exception):
    return error_handler(exception, encode_data)


@app.route('/')
@timed
def index():
    return '''<html>
<head><title>Software Heritage storage server</title></head>
<body>
<p>You have reached the
<a href="https://www.softwareheritage.org/">Software Heritage</a>
storage server.<br />
See its
<a href="https://docs.softwareheritage.org/devel/swh-storage/">documentation
and API</a> for more information</p>
</body>
</html>'''


@app.route('/stat/counters', methods=['GET'])
@timed
def stat_counters():
    return encode_data(get_storage().stat_counters())


@app.route('/stat/refresh', methods=['GET'])
@timed
def refresh_stat_counters():
    return encode_data(get_storage().refresh_stat_counters())


api_cfg = None


def load_and_check_config(config_file, type='local'):
    """Check the minimal configuration is set to run the api or raise an
       error explanation.

    Args:
        config_file (str): Path to the configuration file to load
        type (str): configuration type. For 'local' type, more
                    checks are done.

    Raises:
        Error if the setup is not as expected

    Returns:
        configuration as a dict

    """
    if not config_file:
        raise EnvironmentError('Configuration file must be defined')

    if not os.path.exists(config_file):
        raise FileNotFoundError('Configuration file %s does not exist' % (
            config_file, ))

    cfg = config.read(config_file)
    if 'storage' not in cfg:
        raise KeyError("Missing '%storage' configuration")

    if type == 'local':
        vcfg = cfg['storage']
        cls = vcfg.get('cls')
        if cls != 'local':
            raise ValueError(
                "The storage backend can only be started with a 'local' "
                "configuration")

        args = vcfg['args']
        for key in ('db', 'objstorage'):
            if not args.get(key):
                raise ValueError(
                    "Invalid configuration; missing '%s' config entry" % key)

    return cfg


def make_app_from_configfile():
    """Run the WSGI app from the webserver, loading the configuration from
       a configuration file.

       SWH_CONFIG_FILENAME environment variable defines the
       configuration path to load.

    """
    global api_cfg
    if not api_cfg:
        config_file = os.environ.get('SWH_CONFIG_FILENAME')
        api_cfg = load_and_check_config(config_file)
        app.config.update(api_cfg)
    handler = logging.StreamHandler()
    app.logger.addHandler(handler)
    return app


if __name__ == '__main__':
    print('Deprecated. Use swh-storage')



# Copyright (C) 2019-2020 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging
import psycopg2
import traceback

from datetime import datetime
from typing import Dict, List, Optional, Union

from requests.exceptions import ConnectionError
from tenacity import (
    retry, stop_after_attempt, wait_random_exponential, retry_if_exception_type
)

from swh.storage import get_storage, HashCollision


logger = logging.getLogger(__name__)


RETRY_EXCEPTIONS = [
    # raised when two parallel insertions insert the same data
    psycopg2.IntegrityError,
    HashCollision,
    # when the server is restarting
    ConnectionError,
]


def should_retry_adding(error: Exception) -> bool:
    """Retry if the error/exception if one of the RETRY_EXCEPTIONS type.

    """
    for exc in RETRY_EXCEPTIONS:
        if retry_if_exception_type(exc)(error):
            error_name = error.__module__ + '.' + error.__class__.__name__
            logger.warning('Retry adding a batch', exc_info=False, extra={
                'swh_type': 'storage_retry',
                'swh_exception_type': error_name,
                'swh_exception': traceback.format_exc(),
            })
            return True
    return False


swh_retry = retry(retry=should_retry_adding,
                  wait=wait_random_exponential(multiplier=1, max=10),
                  stop=stop_after_attempt(3))


class RetryingProxyStorage:
    """Storage implementation which retries adding objects when it specifically
       fails (hash collision, integrity error).

    """
    def __init__(self, storage):
        self.storage = get_storage(**storage)

    def __getattr__(self, key):
        return getattr(self.storage, key)

    @swh_retry
    def content_add(self, content: List[Dict]) -> Dict:
        return self.storage.content_add(content)

    @swh_retry
    def content_add_metadata(self, content: List[Dict]) -> Dict:
        return self.storage.content_add_metadata(content)

    @swh_retry
    def origin_add_one(self, origin: Dict) -> str:
        return self.storage.origin_add_one(origin)

    @swh_retry
    def origin_visit_add(self, origin: Dict,
                         date: Union[datetime, str], type: str) -> Dict:
        return self.storage.origin_visit_add(origin, date, type)

    @swh_retry
    def origin_visit_update(
            self, origin: str, visit_id: int, status: Optional[str] = None,
            metadata: Optional[Dict] = None,
            snapshot: Optional[Dict] = None) -> Dict:
        return self.storage.origin_visit_update(
            origin, visit_id, status=status,
            metadata=metadata, snapshot=snapshot)

    @swh_retry
    def tool_add(self, tools: List[Dict]) -> List[Dict]:
        return self.storage.tool_add(tools)

    @swh_retry
    def metadata_provider_add(
            self, provider_name: str, provider_type: str, provider_url: str,
            metadata: Dict) -> Union[str, int]:
        return self.storage.metadata_provider_add(
            provider_name, provider_type, provider_url, metadata)

    @swh_retry
    def origin_metadata_add(
            self, origin_url: str, ts: Union[str, datetime],
            provider_id: int, tool_id: int, metadata: Dict) -> None:
        return self.storage.origin_metadata_add(
            origin_url, ts, provider_id, tool_id, metadata)

    @swh_retry
    def directory_add(self, directories: List[Dict]) -> Dict:
        return self.storage.directory_add(directories)

    @swh_retry
    def revision_add(self, revisions: List[Dict]) -> Dict:
        return self.storage.revision_add(revisions)

    @swh_retry
    def release_add(self, releases: List[Dict]) -> Dict:
        return self.storage.release_add(releases)

    @swh_retry
    def snapshot_add(self, snapshot: List[Dict]) -> Dict:
        return self.storage.snapshot_add(snapshot)



# Copyright (C) 2015-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime
from swh.model.hashutil import hash_to_bytes
from swh.model import from_disk


class StorageData:
    def __getattr__(self, key):
        try:
            v = globals()[key]
        except KeyError as e:
            raise AttributeError(e.args[0])
        if hasattr(v, 'copy'):
            return v.copy()
        return v


data = StorageData()


cont = {
    'data': b'42\n',
    'length': 3,
    'sha1': hash_to_bytes(
        '34973274ccef6ab4dfaaf86599792fa9c3fe4689'),
    'sha1_git': hash_to_bytes(
        'd81cc0710eb6cf9efd5b920a8453e1e07157b6cd'),
    'sha256': hash_to_bytes(
        '673650f936cb3b0a2f93ce09d81be10748b1b203c19e8176b4eefc1964a0cf3a'),
    'blake2s256': hash_to_bytes(
        'd5fe1939576527e42cfd76a9455a2432fe7f56669564577dd93c4280e76d661d'),
    'status': 'visible',
}

cont2 = {
    'data': b'4242\n',
    'length': 5,
    'sha1': hash_to_bytes(
        '61c2b3a30496d329e21af70dd2d7e097046d07b7'),
    'sha1_git': hash_to_bytes(
        '36fade77193cb6d2bd826161a0979d64c28ab4fa'),
    'sha256': hash_to_bytes(
        '859f0b154fdb2d630f45e1ecae4a862915435e663248bb8461d914696fc047cd'),
    'blake2s256': hash_to_bytes(
        '849c20fad132b7c2d62c15de310adfe87be94a379941bed295e8141c6219810d'),
    'status': 'visible',
}

cont3 = {
    'data': b'424242\n',
    'length': 7,
    'sha1': hash_to_bytes(
        '3e21cc4942a4234c9e5edd8a9cacd1670fe59f13'),
    'sha1_git': hash_to_bytes(
        'c932c7649c6dfa4b82327d121215116909eb3bea'),
    'sha256': hash_to_bytes(
        '92fb72daf8c6818288a35137b72155f507e5de8d892712ab96277aaed8cf8a36'),
    'blake2s256': hash_to_bytes(
        '76d0346f44e5a27f6bafdd9c2befd304aff83780f93121d801ab6a1d4769db11'),
    'status': 'visible',
    'ctime': '2019-12-01',
}

contents = (cont, cont2, cont3)


missing_cont = {
    'data': b'missing\n',
    'length': 8,
    'sha1': hash_to_bytes(
        'f9c24e2abb82063a3ba2c44efd2d3c797f28ac90'),
    'sha1_git': hash_to_bytes(
        '33e45d56f88993aae6a0198013efa80716fd8919'),
    'sha256': hash_to_bytes(
        '6bbd052ab054ef222c1c87be60cd191addedd24cc882d1f5f7f7be61dc61bb3a'),
    'blake2s256': hash_to_bytes(
        '306856b8fd879edb7b6f1aeaaf8db9bbecc993cd7f776c333ac3a782fa5c6eba'),
    'status': 'absent',
}

skipped_cont = {
    'length': 1024 * 1024 * 200,
    'sha1_git': hash_to_bytes(
        '33e45d56f88993aae6a0198013efa80716fd8920'),
    'sha1': hash_to_bytes(
        '43e45d56f88993aae6a0198013efa80716fd8920'),
    'sha256': hash_to_bytes(
        '7bbd052ab054ef222c1c87be60cd191addedd24cc882d1f5f7f7be61dc61bb3a'),
    'blake2s256': hash_to_bytes(
        'ade18b1adecb33f891ca36664da676e12c772cc193778aac9a137b8dc5834b9b'),
    'reason': 'Content too long',
    'status': 'absent',
    'origin': 'file:///dev/zero',
}

skipped_cont2 = {
    'length': 1024 * 1024 * 300,
    'sha1_git': hash_to_bytes(
        '44e45d56f88993aae6a0198013efa80716fd8921'),
    'sha1': hash_to_bytes(
        '54e45d56f88993aae6a0198013efa80716fd8920'),
    'sha256': hash_to_bytes(
        '8cbd052ab054ef222c1c87be60cd191addedd24cc882d1f5f7f7be61dc61bb3a'),
    'blake2s256': hash_to_bytes(
        '9ce18b1adecb33f891ca36664da676e12c772cc193778aac9a137b8dc5834b9b'),
    'reason': 'Content too long',
    'status': 'absent',
}

dir = {
    'id': hash_to_bytes(
        '340133423253310030f531e632a733ff37c3a930'),
    'entries': [
        {
            'name': b'foo',
            'type': 'file',
            'target': hash_to_bytes(  # cont
                'd81cc0710eb6cf9efd5b920a8453e1e07157b6cd'),
            'perms': from_disk.DentryPerms.content,
        },
        {
            'name': b'bar\xc3',
            'type': 'dir',
            'target': b'12345678901234567890',
            'perms': from_disk.DentryPerms.directory,
        },
    ],
}

dir2 = {
    'id': hash_to_bytes(
        '340133423253310030f531e632a733ff37c3a935'),
    'entries': [
        {
            'name': b'oof',
            'type': 'file',
            'target': hash_to_bytes(  # cont2
                '36fade77193cb6d2bd826161a0979d64c28ab4fa'),
            'perms': from_disk.DentryPerms.content,
        }
    ],
}

dir3 = {
    'id': hash_to_bytes('33e45d56f88993aae6a0198013efa80716fd8921'),
    'entries': [
        {
            'name': b'foo',
            'type': 'file',
            'target': hash_to_bytes(  # cont
                'd81cc0710eb6cf9efd5b920a8453e1e07157b6cd'),
            'perms': from_disk.DentryPerms.content,
        },
        {
            'name': b'subdir',
            'type': 'dir',
            'target': hash_to_bytes(  # dir
                '340133423253310030f531e632a733ff37c3a930'),
            'perms': from_disk.DentryPerms.directory,
        },
        {
            'name': b'hello',
            'type': 'file',
            'target': b'12345678901234567890',
            'perms': from_disk.DentryPerms.content,
        },
    ],
}

dir4 = {
    'id': hash_to_bytes('33e45d56f88993aae6a0198013efa80716fd8922'),
    'entries': [
        {
            'name': b'subdir1',
            'type': 'dir',
            'target': hash_to_bytes(
                '33e45d56f88993aae6a0198013efa80716fd8921'),  # dir3
            'perms': from_disk.DentryPerms.directory,
        },
    ]
}

dierctories = (dir, dir2, dir3, dir4)


minus_offset = datetime.timezone(datetime.timedelta(minutes=-120))
plus_offset = datetime.timezone(datetime.timedelta(minutes=120))

revision = {
    'id': b'56789012345678901234',
    'message': b'hello',
    'author': {
        'name': b'Nicolas Dandrimont',
        'email': b'nicolas@example.com',
        'fullname': b'Nicolas Dandrimont <nicolas@example.com> ',
    },
    'date': {
        'timestamp': 1234567890,
        'offset': 120,
        'negative_utc': None,
    },
    'committer': {
        'name': b'St\xc3fano Zacchiroli',
        'email': b'stefano@example.com',
        'fullname': b'St\xc3fano Zacchiroli <stefano@example.com>'
    },
    'committer_date': {
        'timestamp': 1123456789,
        'offset': 0,
        'negative_utc': True,
    },
    'parents': [b'01234567890123456789', b'23434512345123456789'],
    'type': 'git',
    'directory': hash_to_bytes(  # dir
        '340133423253310030f531e632a733ff37c3a930'),
    'metadata': {
        'checksums': {
            'sha1': 'tarball-sha1',
            'sha256': 'tarball-sha256',
        },
        'signed-off-by': 'some-dude',
        'extra_headers': [
            ['gpgsig', b'test123'],
            ['mergetags', [b'foo\\bar', b'\x22\xaf\x89\x80\x01\x00']],
        ],
    },
    'synthetic': True
}

revision2 = {
    'id': b'87659012345678904321',
    'message': b'hello again',
    'author': {
        'name': b'Roberto Dicosmo',
        'email': b'roberto@example.com',
        'fullname': b'Roberto Dicosmo <roberto@example.com>',
    },
    'date': {
        'timestamp': {
            'seconds': 1234567843,
            'microseconds': 220000,
        },
        'offset': -720,
        'negative_utc': None,
    },
    'committer': {
        'name': b'tony',
        'email': b'ar@dumont.fr',
        'fullname': b'tony <ar@dumont.fr>',
    },
    'committer_date': {
        'timestamp': 1123456789,
        'offset': 0,
        'negative_utc': False,
    },
    'parents': [b'01234567890123456789'],
    'type': 'git',
    'directory': hash_to_bytes(  # dir2
        '340133423253310030f531e632a733ff37c3a935'),
    'metadata': None,
    'synthetic': False
}

revision3 = {
    'id': hash_to_bytes('7026b7c1a2af56521e951c01ed20f255fa054238'),
    'message': b'a simple revision with no parents this time',
    'author': {
        'name': b'Roberto Dicosmo',
        'email': b'roberto@example.com',
        'fullname': b'Roberto Dicosmo <roberto@example.com>',
    },
    'date': {
        'timestamp': {
            'seconds': 1234567843,
            'microseconds': 220000,
        },
        'offset': -720,
        'negative_utc': None,
    },
    'committer': {
        'name': b'tony',
        'email': b'ar@dumont.fr',
        'fullname': b'tony <ar@dumont.fr>',
    },
    'committer_date': {
        'timestamp': 1127351742,
        'offset': 0,
        'negative_utc': False,
    },
    'parents': [],
    'type': 'git',
    'directory': hash_to_bytes(  # dir2
        '340133423253310030f531e632a733ff37c3a935'),
    'metadata': None,
    'synthetic': True
}

revision4 = {
    'id': hash_to_bytes('368a48fe15b7db2383775f97c6b247011b3f14f4'),
    'message': b'parent of self.revision2',
    'author': {
        'name': b'me',
        'email': b'me@soft.heri',
        'fullname': b'me <me@soft.heri>',
    },
    'date': {
        'timestamp': {
            'seconds': 1244567843,
            'microseconds': 220000,
        },
        'offset': -720,
        'negative_utc': None,
    },
    'committer': {
        'name': b'committer-dude',
        'email': b'committer@dude.com',
        'fullname': b'committer-dude <committer@dude.com>',
    },
    'committer_date': {
        'timestamp': {
            'seconds': 1244567843,
            'microseconds': 220000,
        },
        'offset': -720,
        'negative_utc': None,
    },
    'parents': [hash_to_bytes(  # revision3
        '7026b7c1a2af56521e951c01ed20f255fa054238')],
    'type': 'git',
    'directory': hash_to_bytes(  # dir
        '340133423253310030f531e632a733ff37c3a930'),
    'metadata': None,
    'synthetic': False
}

revisions = (revision, revision2, revision3, revision4)


origin = {
    'url': 'file:///dev/null',
}

origin2 = {
    'url': 'file:///dev/zero',
}

origins = (origin, origin2)


provider = {
    'name': 'hal',
    'type': 'deposit-client',
    'url': 'http:///hal/inria',
    'metadata': {
        'location': 'France'
    }
}

metadata_tool = {
    'name': 'swh-deposit',
    'version': '0.0.1',
    'configuration': {
        'sword_version': '2'
    }
}

date_visit1 = datetime.datetime(2015, 1, 1, 23, 0, 0,
                                tzinfo=datetime.timezone.utc)
type_visit1 = 'git'

date_visit2 = datetime.datetime(2017, 1, 1, 23, 0, 0,
                                tzinfo=datetime.timezone.utc)
type_visit2 = 'hg'

date_visit3 = datetime.datetime(2018, 1, 1, 23, 0, 0,
                                tzinfo=datetime.timezone.utc)
type_visit3 = 'deb'

release = {
    'id': b'87659012345678901234',
    'name': b'v0.0.1',
    'author': {
        'name': b'olasd',
        'email': b'nic@olasd.fr',
        'fullname': b'olasd <nic@olasd.fr>',
    },
    'date': {
        'timestamp': 1234567890,
        'offset': 42,
        'negative_utc': None,
    },
    'target': b'43210987654321098765',
    'target_type': 'revision',
    'message': b'synthetic release',
    'synthetic': True,
}

release2 = {
    'id': b'56789012348765901234',
    'name': b'v0.0.2',
    'author': {
        'name': b'tony',
        'email': b'ar@dumont.fr',
        'fullname': b'tony <ar@dumont.fr>',
    },
    'date': {
        'timestamp': 1634366813,
        'offset': -120,
        'negative_utc': None,
    },
    'target': b'432109\xa9765432\xc309\x00765',
    'target_type': 'revision',
    'message': b'v0.0.2\nMisc performance improvements + bug fixes',
    'synthetic': False
}

release3 = {
    'id': b'87659012345678904321',
    'name': b'v0.0.2',
    'author': {
        'name': b'tony',
        'email': b'tony@ardumont.fr',
        'fullname': b'tony <tony@ardumont.fr>',
    },
    'date': {
        'timestamp': 1634336813,
        'offset': 0,
        'negative_utc': False,
    },
    'target': b'87659012345678904321',  # revision2
    'target_type': 'revision',
    'message': b'yet another synthetic release',
    'synthetic': True,
}

releases = (release, release2, release3)


snapshot = {
    'id': hash_to_bytes('2498dbf535f882bc7f9a18fb16c9ad27fda7bab7'),
    'branches': {
        b'master': {
            'target': b'56789012345678901234',  # revision
            'target_type': 'revision',
        },
    },
}

empty_snapshot = {
    'id': hash_to_bytes('1a8893e6a86f444e8be8e7bda6cb34fb1735a00e'),
    'branches': {},
}

complete_snapshot = {
    'id': hash_to_bytes('6e65b86363953b780d92b0a928f3e8fcdd10db36'),
    'branches': {
        b'directory': {
            'target': hash_to_bytes(
                '1bd0e65f7d2ff14ae994de17a1e7fe65111dcad8'),
            'target_type': 'directory',
        },
        b'directory2': {
            'target': hash_to_bytes(
                '1bd0e65f7d2ff14ae994de17a1e7fe65111dcad8'),
            'target_type': 'directory',
        },
        b'content': {
            'target': hash_to_bytes(
                'fe95a46679d128ff167b7c55df5d02356c5a1ae1'),
            'target_type': 'content',
        },
        b'alias': {
            'target': b'revision',
            'target_type': 'alias',
        },
        b'revision': {
            'target': hash_to_bytes(
                'aafb16d69fd30ff58afdd69036a26047f3aebdc6'),
            'target_type': 'revision',
        },
        b'release': {
            'target': hash_to_bytes(
                '7045404f3d1c54e6473c71bbb716529fbad4be24'),
            'target_type': 'release',
        },
        b'snapshot': {
            'target': hash_to_bytes(
                '1a8893e6a86f444e8be8e7bda6cb34fb1735a00e'),
            'target_type': 'snapshot',
        },
        b'dangling': None,
    }
}

origin_metadata = {
        'origin': origin,
        'discovery_date': datetime.datetime(2015, 1, 1, 23, 0, 0,
                                            tzinfo=datetime.timezone.utc),
        'provider': provider,
        'tool': 'swh-deposit',
        'metadata': {
            'name': 'test_origin_metadata',
            'version': '0.0.1'
        }
    }
origin_metadata2 = {
        'origin': origin,
        'discovery_date': datetime.datetime(2017, 1, 1, 23, 0, 0,
                                            tzinfo=datetime.timezone.utc),
        'provider': provider,
        'tool': 'swh-deposit',
        'metadata': {
            'name': 'test_origin_metadata',
            'version': '0.0.1'
        }
    }

person = {
    'name': b'John Doe',
    'email': b'john.doe@institute.org',
    'fullname': b'John Doe <john.doe@institute.org>'
}



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from unittest.mock import patch

from swh.storage.in_memory import Storage
from swh.storage.algos.origin import iter_origins


def assert_list_eq(left, right, msg=None):
    assert list(left) == list(right), msg


def test_iter_origins():
    storage = Storage()
    origins = storage.origin_add([
        {'url': 'bar'},
        {'url': 'qux'},
        {'url': 'quuz'},
    ])
    assert_list_eq(iter_origins(storage), origins)
    assert_list_eq(iter_origins(storage, batch_size=1), origins)
    assert_list_eq(iter_origins(storage, batch_size=2), origins)

    for i in range(1, 5):
        assert_list_eq(
            iter_origins(storage, origin_from=i+1),
            origins[i:],
            i)

        assert_list_eq(
            iter_origins(storage, origin_from=i+1, batch_size=1),
            origins[i:],
            i)

        assert_list_eq(
            iter_origins(storage, origin_from=i+1, batch_size=2),
            origins[i:],
            i)

        for j in range(i, 5):
            assert_list_eq(
                iter_origins(
                    storage, origin_from=i+1, origin_to=j+1),
                origins[i:j],
                (i, j))

            assert_list_eq(
                iter_origins(
                    storage, origin_from=i+1, origin_to=j+1, batch_size=1),
                origins[i:j],
                (i, j))

            assert_list_eq(
                iter_origins(
                    storage, origin_from=i+1, origin_to=j+1, batch_size=2),
                origins[i:j],
                (i, j))


@patch('swh.storage.in_memory.Storage.origin_get_range')
def test_iter_origins_batch_size(mock_origin_get_range):
    storage = Storage()
    mock_origin_get_range.return_value = []

    list(iter_origins(storage))
    mock_origin_get_range.assert_called_with(
        origin_from=1, origin_count=10000)

    list(iter_origins(storage, batch_size=42))
    mock_origin_get_range.assert_called_with(
        origin_from=1, origin_count=42)



# Copyright (C) 2015-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest

from swh.storage.api.client import RemoteStorage
import swh.storage.api.server as server
import swh.storage.storage
from swh.storage.tests.test_storage import (  # noqa
    TestStorage, TestStorageGeneratedData)

# tests are executed using imported classes (TestStorage and
# TestStorageGeneratedData) using overloaded swh_storage fixture
# below


@pytest.fixture
def app():
    storage_config = {
        'cls': 'memory',
        'journal_writer': {
            'cls': 'memory',
        },
    }
    server.storage = swh.storage.get_storage(**storage_config)
    # hack hack hack!
    # We attach the journal storage to the app here to make it accessible to
    # the test (as swh_storage.journal_writer); see swh_storage below.
    server.app.journal_writer = server.storage.journal_writer
    yield server.app
    del server.app.journal_writer


@pytest.fixture
def swh_rpc_client_class():
    return RemoteStorage


@pytest.fixture
def swh_storage(swh_rpc_client, app):
    # This version of the swh_storage fixture uses the swh_rpc_client fixture
    # to instantiate a RemoteStorage (see swh_rpc_client_class above) that
    # proxies, via the swh.core RPC mechanism, the local (in memory) storage
    # configured in the app fixture above.
    #
    # Also note that, for the sake of
    # making it easier to write tests, the in-memory journal writer of the
    # in-memory backend storage is attached to the RemoteStorage as its
    # journal_writer attribute.
    storage = swh_rpc_client
    journal_writer = getattr(storage, 'journal_writer', None)
    storage.journal_writer = app.journal_writer
    yield storage
    storage.journal_writer = journal_writer



# Copyright (C) 2015-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import re
import bisect
import dateutil
import collections
import copy
import datetime
import itertools
import random

from collections import defaultdict
from datetime import timedelta
from typing import Any, Dict, List, Mapping, Optional

import attr

from swh.model.model import (
    Content, Directory, Revision, Release, Snapshot, OriginVisit, Origin,
    SHA1_SIZE)
from swh.model.hashutil import DEFAULT_ALGORITHMS, hash_to_bytes, hash_to_hex
from swh.objstorage import get_objstorage
from swh.objstorage.exc import ObjNotFoundError

from .storage import get_journal_writer
from .converters import origin_url_to_sha1
from .utils import get_partition_bounds_bytes

# Max block size of contents to return
BULK_BLOCK_CONTENT_LEN_MAX = 10000


def now():
    return datetime.datetime.now(tz=datetime.timezone.utc)


class Storage:
    def __init__(self, journal_writer=None):
        self._contents = {}
        self._content_indexes = defaultdict(lambda: defaultdict(set))
        self._skipped_contents = {}
        self._skipped_content_indexes = defaultdict(lambda: defaultdict(set))

        self.reset()

        if journal_writer:
            self.journal_writer = get_journal_writer(**journal_writer)
        else:
            self.journal_writer = None

    def reset(self):
        self._directories = {}
        self._revisions = {}
        self._releases = {}
        self._snapshots = {}
        self._origins = {}
        self._origins_by_id = []
        self._origins_by_sha1 = {}
        self._origin_visits = {}
        self._persons = []
        self._origin_metadata = defaultdict(list)
        self._tools = {}
        self._metadata_providers = {}
        self._objects = defaultdict(list)

        # ideally we would want a skip list for both fast inserts and searches
        self._sorted_sha1s = []

        self.objstorage = get_objstorage('memory', {})

    def check_config(self, *, check_write):
        """Check that the storage is configured and ready to go."""
        return True

    def _content_add(self, contents, with_data):
        content_with_data = []
        content_without_data = []
        for content in contents:
            if content.status is None:
                content.status = 'visible'
            if content.length is None:
                content.length = -1
            if content.status != 'absent':
                if self._content_key(content) not in self._contents:
                    content_with_data.append(content)
            else:
                if self._content_key(content) not in self._skipped_contents:
                    content_without_data.append(content)

        if self.journal_writer:
            for content in content_with_data:
                content = attr.evolve(content, data=None)
                self.journal_writer.write_addition('content', content)
            for content in content_without_data:
                self.journal_writer.write_addition('content', content)

        count_content_added, count_content_bytes_added = \
            self._content_add_present(content_with_data, with_data)

        count_skipped_content_added = self._content_add_absent(
            content_without_data
        )

        summary = {
            'content:add': count_content_added,
            'skipped_content:add': count_skipped_content_added,
        }

        if with_data:
            summary['content:add:bytes'] = count_content_bytes_added

        return summary

    def _content_add_present(self, contents, with_data):
        count_content_added = 0
        count_content_bytes_added = 0
        for content in contents:
            key = self._content_key(content)
            if key in self._contents:
                continue
            for algorithm in DEFAULT_ALGORITHMS:
                hash_ = content.get_hash(algorithm)
                if hash_ in self._content_indexes[algorithm]\
                   and (algorithm not in {'blake2s256', 'sha256'}):
                    from . import HashCollision
                    raise HashCollision(algorithm, hash_, key)
            for algorithm in DEFAULT_ALGORITHMS:
                hash_ = content.get_hash(algorithm)
                self._content_indexes[algorithm][hash_].add(key)
            self._objects[content.sha1_git].append(
                ('content', content.sha1))
            self._contents[key] = content
            bisect.insort(self._sorted_sha1s, content.sha1)
            count_content_added += 1
            if with_data:
                content_data = self._contents[key].data
                self._contents[key] = attr.evolve(
                    self._contents[key],
                    data=None)
                count_content_bytes_added += len(content_data)
                self.objstorage.add(content_data, content.sha1)

        return (count_content_added, count_content_bytes_added)

    def _content_add_absent(self, contents):
        count = 0
        skipped_content_missing = self.skipped_content_missing(contents)
        for content in skipped_content_missing:
            key = self._content_key(content)
            for algo in DEFAULT_ALGORITHMS:
                self._skipped_content_indexes[algo][content.get_hash(algo)] \
                    .add(key)
            self._skipped_contents[key] = content
            count += 1

        return count

    def _content_to_model(self, contents):
        """Takes a list of content dicts, optionally with an extra 'origin'
        key, and yields tuples (model.Content, origin)."""
        for content in contents:
            content = content.copy()
            content.pop('origin', None)
            yield Content.from_dict(content)

    def content_add(self, content):
        """Add content blobs to the storage

        Args:
            content (iterable): iterable of dictionaries representing
                individual pieces of content to add. Each dictionary has the
                following keys:

                - data (bytes): the actual content
                - length (int): content length (default: -1)
                - one key for each checksum algorithm in
                  :data:`swh.model.hashutil.DEFAULT_ALGORITHMS`, mapped to the
                  corresponding checksum
                - status (str): one of visible, hidden, absent
                - reason (str): if status = absent, the reason why
                - origin (int): if status = absent, the origin we saw the
                  content in

        Raises:
            HashCollision in case of collision

        Returns:
            Summary dict with the following key and associated values:

                content:add: New contents added
                content_bytes:add: Sum of the contents' length data
                skipped_content:add: New skipped contents (no data) added

        """
        now = datetime.datetime.now(tz=datetime.timezone.utc)
        content = [attr.evolve(c, ctime=now)
                   for c in self._content_to_model(content)]
        return self._content_add(content, with_data=True)

    def content_add_metadata(self, content):
        """Add content metadata to the storage (like `content_add`, but
        without inserting to the objstorage).

        Args:
            content (iterable): iterable of dictionaries representing
                individual pieces of content to add. Each dictionary has the
                following keys:

                - length (int): content length (default: -1)
                - one key for each checksum algorithm in
                  :data:`swh.model.hashutil.DEFAULT_ALGORITHMS`, mapped to the
                  corresponding checksum
                - status (str): one of visible, hidden, absent
                - reason (str): if status = absent, the reason why
                - origin (int): if status = absent, the origin we saw the
                  content in
                - ctime (datetime): time of insertion in the archive

        Raises:
            HashCollision in case of collision

        Returns:
            Summary dict with the following key and associated values:

                content:add: New contents added
                skipped_content:add: New skipped contents (no data) added

        """
        content = list(self._content_to_model(content))
        return self._content_add(content, with_data=False)

    def content_get(self, content):
        """Retrieve in bulk contents and their data.

        This function may yield more blobs than provided sha1 identifiers,
        in case they collide.

        Args:
            content: iterables of sha1

        Yields:
            Dict[str, bytes]: Generates streams of contents as dict with their
                raw data:

                - sha1 (bytes): content id
                - data (bytes): content's raw data

        Raises:
            ValueError in case of too much contents are required.
            cf. BULK_BLOCK_CONTENT_LEN_MAX

        """
        # FIXME: Make this method support slicing the `data`.
        if len(content) > BULK_BLOCK_CONTENT_LEN_MAX:
            raise ValueError(
                "Sending at most %s contents." % BULK_BLOCK_CONTENT_LEN_MAX)
        for obj_id in content:
            try:
                data = self.objstorage.get(obj_id)
            except ObjNotFoundError:
                yield None
                continue

            yield {'sha1': obj_id, 'data': data}

    def content_get_range(self, start, end, limit=1000, db=None, cur=None):
        """Retrieve contents within range [start, end] bound by limit.

        Note that this function may return more than one blob per hash. The
        limit is enforced with multiplicity (ie. two blobs with the same hash
        will count twice toward the limit).

        Args:
            **start** (bytes): Starting identifier range (expected smaller
                           than end)
            **end** (bytes): Ending identifier range (expected larger
                             than start)
            **limit** (int): Limit result (default to 1000)

        Returns:
            a dict with keys:
            - contents [dict]: iterable of contents in between the range.
            - next (bytes): There remains content in the range
              starting from this next sha1

        """
        if limit is None:
            raise ValueError('Development error: limit should not be None')
        from_index = bisect.bisect_left(self._sorted_sha1s, start)
        sha1s = itertools.islice(self._sorted_sha1s, from_index, None)
        sha1s = ((sha1, content_key)
                 for sha1 in sha1s
                 for content_key in self._content_indexes['sha1'][sha1])
        matched = []
        next_content = None
        for sha1, key in sha1s:
            if sha1 > end:
                break
            if len(matched) >= limit:
                next_content = sha1
                break
            matched.append(self._contents[key].to_dict())
        return {
            'contents': matched,
            'next': next_content,
        }

    def content_get_partition(
            self, partition_id: int, nb_partitions: int, limit: int = 1000,
            page_token: str = None):
        """Splits contents into nb_partitions, and returns one of these based on
        partition_id (which must be in [0, nb_partitions-1])

        There is no guarantee on how the partitioning is done, or the
        result order.

        Args:
            partition_id (int): index of the partition to fetch
            nb_partitions (int): total number of partitions to split into
            limit (int): Limit result (default to 1000)
            page_token (Optional[str]): opaque token used for pagination.

        Returns:
            a dict with keys:
              - contents (List[dict]): iterable of contents in the partition.
              - **next_page_token** (Optional[str]): opaque token to be used as
                `page_token` for retrieving the next page. if absent, there is
                no more pages to gather.
        """
        if limit is None:
            raise ValueError('Development error: limit should not be None')
        (start, end) = get_partition_bounds_bytes(
            partition_id, nb_partitions, SHA1_SIZE)
        if page_token:
            start = hash_to_bytes(page_token)
        if end is None:
            end = b'\xff'*SHA1_SIZE
        result = self.content_get_range(start, end, limit)
        result2 = {
            'contents': result['contents'],
            'next_page_token': None,
        }
        if result['next']:
            result2['next_page_token'] = hash_to_hex(result['next'])
        return result2

    def content_get_metadata(
            self, contents: List[bytes]) -> Dict[bytes, List[Dict]]:
        """Retrieve content metadata in bulk

        Args:
            content: iterable of content identifiers (sha1)

        Returns:
            a dict with keys the content's sha1 and the associated value
            either the existing content's metadata or None if the content does
            not exist.

        """
        result: Dict = {sha1: [] for sha1 in contents}
        for sha1 in contents:
            if sha1 in self._content_indexes['sha1']:
                objs = self._content_indexes['sha1'][sha1]
                # only 1 element as content_add_metadata would have raised a
                # hash collision otherwise
                for key in objs:
                    d = self._contents[key].to_dict()
                    del d['ctime']
                    result[sha1].append(d)
        return result

    def content_find(self, content):
        if not set(content).intersection(DEFAULT_ALGORITHMS):
            raise ValueError('content keys must contain at least one of: '
                             '%s' % ', '.join(sorted(DEFAULT_ALGORITHMS)))
        found = []
        for algo in DEFAULT_ALGORITHMS:
            hash = content.get(algo)
            if hash and hash in self._content_indexes[algo]:
                found.append(self._content_indexes[algo][hash])

        if not found:
            return []

        keys = list(set.intersection(*found))
        return [self._contents[key].to_dict() for key in keys]

    def content_missing(self, content, key_hash='sha1'):
        """List content missing from storage

        Args:
            contents ([dict]): iterable of dictionaries whose keys are
                               either 'length' or an item of
                               :data:`swh.model.hashutil.ALGORITHMS`;
                               mapped to the corresponding checksum
                               (or length).

            key_hash (str): name of the column to use as hash id
                            result (default: 'sha1')

        Returns:
            iterable ([bytes]): missing content ids (as per the
            key_hash column)
        """
        for cont in content:
            for (algo, hash_) in cont.items():
                if algo not in DEFAULT_ALGORITHMS:
                    continue
                if hash_ not in self._content_indexes.get(algo, []):
                    yield cont[key_hash]
                    break
            else:
                for result in self.content_find(cont):
                    if result['status'] == 'missing':
                        yield cont[key_hash]

    def content_missing_per_sha1(self, contents):
        """List content missing from storage based only on sha1.

        Args:
            contents: Iterable of sha1 to check for absence.

        Returns:
            iterable: missing ids

        Raises:
            TODO: an exception when we get a hash collision.

        """
        for content in contents:
            if content not in self._content_indexes['sha1']:
                yield content

    def skipped_content_missing(self, contents):
        """List all skipped_content missing from storage

        Args:
            contents: Iterable of sha1 to check for skipped content entry

        Returns:
            iterable: dict of skipped content entry
        """

        for content in contents:
            for (key, algorithm) in self._content_key_algorithm(content):
                if algorithm == 'blake2s256':
                    continue
                if key not in self._skipped_content_indexes[algorithm]:
                    # index must contain hashes of algos except blake2s256
                    # else the content is considered skipped
                    yield content
                    break

    def content_get_random(self):
        """Finds a random content id.

        Returns:
            a sha1_git
        """
        return random.choice(list(self._content_indexes['sha1_git']))

    def directory_add(self, directories):
        """Add directories to the storage

        Args:
            directories (iterable): iterable of dictionaries representing the
                individual directories to add. Each dict has the following
                keys:

                - id (sha1_git): the id of the directory to add
                - entries (list): list of dicts for each entry in the
                      directory.  Each dict has the following keys:

                      - name (bytes)
                      - type (one of 'file', 'dir', 'rev'): type of the
                        directory entry (file, directory, revision)
                      - target (sha1_git): id of the object pointed at by the
                        directory entry
                      - perms (int): entry permissions
        Returns:
            Summary dict of keys with associated count as values:

                directory:add: Number of directories actually added

        """
        if self.journal_writer:
            self.journal_writer.write_additions(
                'directory',
                (dir_ for dir_ in directories
                 if dir_['id'] not in self._directories))

        directories = [Directory.from_dict(d) for d in directories]

        count = 0
        for directory in directories:
            if directory.id not in self._directories:
                count += 1
                self._directories[directory.id] = directory
                self._objects[directory.id].append(
                    ('directory', directory.id))

        return {'directory:add': count}

    def directory_missing(self, directories):
        """List directories missing from storage

        Args:
            directories (iterable): an iterable of directory ids

        Yields:
            missing directory ids

        """
        for id in directories:
            if id not in self._directories:
                yield id

    def _join_dentry_to_content(self, dentry):
        keys = (
            'status',
            'sha1',
            'sha1_git',
            'sha256',
            'length',
        )
        ret = dict.fromkeys(keys)
        ret.update(dentry)
        if ret['type'] == 'file':
            # TODO: Make it able to handle more than one content
            content = self.content_find({'sha1_git': ret['target']})
            if content:
                content = content[0]
                for key in keys:
                    ret[key] = content[key]
        return ret

    def _directory_ls(self, directory_id, recursive, prefix=b''):
        if directory_id in self._directories:
            for entry in self._directories[directory_id].entries:
                ret = self._join_dentry_to_content(entry.to_dict())
                ret['name'] = prefix + ret['name']
                ret['dir_id'] = directory_id
                yield ret
                if recursive and ret['type'] == 'dir':
                    yield from self._directory_ls(
                        ret['target'], True, prefix + ret['name'] + b'/')

    def directory_ls(self, directory, recursive=False):
        """Get entries for one directory.

        Args:
            - directory: the directory to list entries from.
            - recursive: if flag on, this list recursively from this directory.

        Returns:
            List of entries for such directory.

        If `recursive=True`, names in the path of a dir/file not at the
        root are concatenated with a slash (`/`).
        """
        yield from self._directory_ls(directory, recursive)

    def directory_entry_get_by_path(self, directory, paths):
        """Get the directory entry (either file or dir) from directory with path.

        Args:
            - directory: sha1 of the top level directory
            - paths: path to lookup from the top level directory. From left
              (top) to right (bottom).

        Returns:
            The corresponding directory entry if found, None otherwise.

        """
        return self._directory_entry_get_by_path(directory, paths, b'')

    def directory_get_random(self):
        """Finds a random directory id.

        Returns:
            a sha1_git if any

        """
        if not self._directories:
            return None
        return random.choice(list(self._directories))

    def _directory_entry_get_by_path(self, directory, paths, prefix):
        if not paths:
            return

        contents = list(self.directory_ls(directory))

        if not contents:
            return

        def _get_entry(entries, name):
            for entry in entries:
                if entry['name'] == name:
                    entry = entry.copy()
                    entry['name'] = prefix + entry['name']
                    return entry

        first_item = _get_entry(contents, paths[0])

        if len(paths) == 1:
            return first_item

        if not first_item or first_item['type'] != 'dir':
            return

        return self._directory_entry_get_by_path(
                first_item['target'], paths[1:], prefix + paths[0] + b'/')

    def revision_add(self, revisions):
        """Add revisions to the storage

        Args:
            revisions (Iterable[dict]): iterable of dictionaries representing
                the individual revisions to add. Each dict has the following
                keys:

                - **id** (:class:`sha1_git`): id of the revision to add
                - **date** (:class:`dict`): date the revision was written
                - **committer_date** (:class:`dict`): date the revision got
                  added to the origin
                - **type** (one of 'git', 'tar'): type of the
                  revision added
                - **directory** (:class:`sha1_git`): the directory the
                  revision points at
                - **message** (:class:`bytes`): the message associated with
                  the revision
                - **author** (:class:`Dict[str, bytes]`): dictionary with
                  keys: name, fullname, email
                - **committer** (:class:`Dict[str, bytes]`): dictionary with
                  keys: name, fullname, email
                - **metadata** (:class:`jsonb`): extra information as
                  dictionary
                - **synthetic** (:class:`bool`): revision's nature (tarball,
                  directory creates synthetic revision`)
                - **parents** (:class:`list[sha1_git]`): the parents of
                  this revision

        date dictionaries have the form defined in :mod:`swh.model`.

        Returns:
            Summary dict of keys with associated count as values

                revision_added: New objects actually stored in db

        """
        if self.journal_writer:
            self.journal_writer.write_additions(
                'revision',
                (rev for rev in revisions
                 if rev['id'] not in self._revisions))

        revisions = [Revision.from_dict(rev) for rev in revisions]

        count = 0
        for revision in revisions:
            if revision.id not in self._revisions:
                revision = attr.evolve(
                    revision,
                    committer=self._person_add(revision.committer),
                    author=self._person_add(revision.author))
                self._revisions[revision.id] = revision
                self._objects[revision.id].append(
                    ('revision', revision.id))
                count += 1

        return {'revision:add': count}

    def revision_missing(self, revisions):
        """List revisions missing from storage

        Args:
            revisions (iterable): revision ids

        Yields:
            missing revision ids

        """
        for id in revisions:
            if id not in self._revisions:
                yield id

    def revision_get(self, revisions):
        for id in revisions:
            if id in self._revisions:
                yield self._revisions.get(id).to_dict()
            else:
                yield None

    def _get_parent_revs(self, rev_id, seen, limit):
        if limit and len(seen) >= limit:
            return
        if rev_id in seen or rev_id not in self._revisions:
            return
        seen.add(rev_id)
        yield self._revisions[rev_id].to_dict()
        for parent in self._revisions[rev_id].parents:
            yield from self._get_parent_revs(parent, seen, limit)

    def revision_log(self, revisions, limit=None):
        """Fetch revision entry from the given root revisions.

        Args:
            revisions: array of root revision to lookup
            limit: limitation on the output result. Default to None.

        Yields:
            List of revision log from such revisions root.

        """
        seen = set()
        for rev_id in revisions:
            yield from self._get_parent_revs(rev_id, seen, limit)

    def revision_shortlog(self, revisions, limit=None):
        """Fetch the shortlog for the given revisions

        Args:
            revisions: list of root revisions to lookup
            limit: depth limitation for the output

        Yields:
            a list of (id, parents) tuples.

        """
        yield from ((rev['id'], rev['parents'])
                    for rev in self.revision_log(revisions, limit))

    def revision_get_random(self):
        """Finds a random revision id.

        Returns:
            a sha1_git
        """
        return random.choice(list(self._revisions))

    def release_add(self, releases):
        """Add releases to the storage

        Args:
            releases (Iterable[dict]): iterable of dictionaries representing
                the individual releases to add. Each dict has the following
                keys:

                - **id** (:class:`sha1_git`): id of the release to add
                - **revision** (:class:`sha1_git`): id of the revision the
                  release points to
                - **date** (:class:`dict`): the date the release was made
                - **name** (:class:`bytes`): the name of the release
                - **comment** (:class:`bytes`): the comment associated with
                  the release
                - **author** (:class:`Dict[str, bytes]`): dictionary with
                  keys: name, fullname, email

        the date dictionary has the form defined in :mod:`swh.model`.

        Returns:
            Summary dict of keys with associated count as values

                release:add: New objects contents actually stored in db

        """
        if self.journal_writer:
            self.journal_writer.write_additions(
                'release',
                (rel for rel in releases
                 if rel['id'] not in self._releases))

        releases = [Release.from_dict(rel) for rel in releases]

        count = 0
        for rel in releases:
            if rel.id not in self._releases:
                if rel.author:
                    self._person_add(rel.author)
                self._objects[rel.id].append(
                    ('release', rel.id))
                self._releases[rel.id] = rel
                count += 1

        return {'release:add': count}

    def release_missing(self, releases):
        """List releases missing from storage

        Args:
            releases: an iterable of release ids

        Returns:
            a list of missing release ids

        """
        yield from (rel for rel in releases if rel not in self._releases)

    def release_get(self, releases):
        """Given a list of sha1, return the releases's information

        Args:
            releases: list of sha1s

        Yields:
            dicts with the same keys as those given to `release_add`
            (or ``None`` if a release does not exist)

        """
        for rel_id in releases:
            if rel_id in self._releases:
                yield self._releases[rel_id].to_dict()
            else:
                yield None

    def release_get_random(self):
        """Finds a random release id.

        Returns:
            a sha1_git
        """
        return random.choice(list(self._releases))

    def snapshot_add(self, snapshots):
        """Add a snapshot to the storage

        Args:
            snapshot ([dict]): the snapshots to add, containing the
              following keys:

              - **id** (:class:`bytes`): id of the snapshot
              - **branches** (:class:`dict`): branches the snapshot contains,
                mapping the branch name (:class:`bytes`) to the branch target,
                itself a :class:`dict` (or ``None`` if the branch points to an
                unknown object)

                - **target_type** (:class:`str`): one of ``content``,
                  ``directory``, ``revision``, ``release``,
                  ``snapshot``, ``alias``
                - **target** (:class:`bytes`): identifier of the target
                  (currently a ``sha1_git`` for all object kinds, or the name
                  of the target branch for aliases)

        Raises:
            ValueError: if the origin's or visit's identifier does not exist.

        Returns:
            Summary dict of keys with associated count as values

                snapshot_added: Count of object actually stored in db

        """
        count = 0
        snapshots = (Snapshot.from_dict(d) for d in snapshots)
        snapshots = (snap for snap in snapshots
                     if snap.id not in self._snapshots)
        for snapshot in snapshots:
            if self.journal_writer:
                self.journal_writer.write_addition('snapshot', snapshot)

            sorted_branch_names = sorted(snapshot.branches)
            self._snapshots[snapshot.id] = (snapshot, sorted_branch_names)
            self._objects[snapshot.id].append(('snapshot', snapshot.id))
            count += 1

        return {'snapshot:add': count}

    def snapshot_get(self, snapshot_id):
        """Get the content, possibly partial, of a snapshot with the given id

        The branches of the snapshot are iterated in the lexicographical
        order of their names.

        .. warning:: At most 1000 branches contained in the snapshot will be
            returned for performance reasons. In order to browse the whole
            set of branches, the method :meth:`snapshot_get_branches`
            should be used instead.

        Args:
            snapshot_id (bytes): identifier of the snapshot
        Returns:
            dict: a dict with three keys:
                * **id**: identifier of the snapshot
                * **branches**: a dict of branches contained in the snapshot
                  whose keys are the branches' names.
                * **next_branch**: the name of the first branch not returned
                  or :const:`None` if the snapshot has less than 1000
                  branches.
        """
        return self.snapshot_get_branches(snapshot_id)

    def snapshot_get_by_origin_visit(self, origin, visit):
        """Get the content, possibly partial, of a snapshot for the given origin visit

        The branches of the snapshot are iterated in the lexicographical
        order of their names.

        .. warning:: At most 1000 branches contained in the snapshot will be
            returned for performance reasons. In order to browse the whole
            set of branches, the method :meth:`snapshot_get_branches`
            should be used instead.

        Args:
            origin (int): the origin's identifier
            visit (int): the visit's identifier
        Returns:
            dict: None if the snapshot does not exist;
              a dict with three keys otherwise:
                * **id**: identifier of the snapshot
                * **branches**: a dict of branches contained in the snapshot
                  whose keys are the branches' names.
                * **next_branch**: the name of the first branch not returned
                  or :const:`None` if the snapshot has less than 1000
                  branches.

        """
        origin_url = self._get_origin_url(origin)
        if not origin_url:
            return

        if origin_url not in self._origins or \
           visit > len(self._origin_visits[origin_url]):
            return None
        snapshot_id = self._origin_visits[origin_url][visit-1].snapshot
        if snapshot_id:
            return self.snapshot_get(snapshot_id)
        else:
            return None

    def snapshot_get_latest(self, origin, allowed_statuses=None):
        """Get the content, possibly partial, of the latest snapshot for the
        given origin, optionally only from visits that have one of the given
        allowed_statuses

        The branches of the snapshot are iterated in the lexicographical
        order of their names.

        .. warning:: At most 1000 branches contained in the snapshot will be
            returned for performance reasons. In order to browse the whole
            set of branches, the methods :meth:`origin_visit_get_latest`
            and :meth:`snapshot_get_branches` should be used instead.

        Args:
            origin (str): the origin's URL
            allowed_statuses (list of str): list of visit statuses considered
                to find the latest snapshot for the origin. For instance,
                ``allowed_statuses=['full']`` will only consider visits that
                have successfully run to completion.
        Returns:
            dict: a dict with three keys:
                * **id**: identifier of the snapshot
                * **branches**: a dict of branches contained in the snapshot
                  whose keys are the branches' names.
                * **next_branch**: the name of the first branch not returned
                  or :const:`None` if the snapshot has less than 1000
                  branches.
        """
        origin_url = self._get_origin_url(origin)
        if not origin_url:
            return

        visit = self.origin_visit_get_latest(
            origin_url,
            allowed_statuses=allowed_statuses,
            require_snapshot=True)
        if visit and visit['snapshot']:
            snapshot = self.snapshot_get(visit['snapshot'])
            if not snapshot:
                raise ValueError(
                    'last origin visit references an unknown snapshot')
            return snapshot

    def snapshot_count_branches(self, snapshot_id, db=None, cur=None):
        """Count the number of branches in the snapshot with the given id

        Args:
            snapshot_id (bytes): identifier of the snapshot

        Returns:
            dict: A dict whose keys are the target types of branches and
            values their corresponding amount
        """
        (snapshot, _) = self._snapshots[snapshot_id]
        return collections.Counter(branch.target_type.value if branch else None
                                   for branch in snapshot.branches.values())

    def snapshot_get_branches(self, snapshot_id, branches_from=b'',
                              branches_count=1000, target_types=None):
        """Get the content, possibly partial, of a snapshot with the given id

        The branches of the snapshot are iterated in the lexicographical
        order of their names.

        Args:
            snapshot_id (bytes): identifier of the snapshot
            branches_from (bytes): optional parameter used to skip branches
                whose name is lesser than it before returning them
            branches_count (int): optional parameter used to restrain
                the amount of returned branches
            target_types (list): optional parameter used to filter the
                target types of branch to return (possible values that can be
                contained in that list are `'content', 'directory',
                'revision', 'release', 'snapshot', 'alias'`)
        Returns:
            dict: None if the snapshot does not exist;
              a dict with three keys otherwise:
                * **id**: identifier of the snapshot
                * **branches**: a dict of branches contained in the snapshot
                  whose keys are the branches' names.
                * **next_branch**: the name of the first branch not returned
                  or :const:`None` if the snapshot has less than
                  `branches_count` branches after `branches_from` included.
        """
        res = self._snapshots.get(snapshot_id)
        if res is None:
            return None
        (snapshot, sorted_branch_names) = res
        from_index = bisect.bisect_left(
                sorted_branch_names, branches_from)
        if target_types:
            next_branch = None
            branches = {}
            for branch_name in sorted_branch_names[from_index:]:
                branch = snapshot.branches[branch_name]
                if branch and branch.target_type.value in target_types:
                    if len(branches) < branches_count:
                        branches[branch_name] = branch
                    else:
                        next_branch = branch_name
                        break
        else:
            # As there is no 'target_types', we can do that much faster
            to_index = from_index + branches_count
            returned_branch_names = sorted_branch_names[from_index:to_index]
            branches = {branch_name: snapshot.branches[branch_name]
                        for branch_name in returned_branch_names}
            if to_index >= len(sorted_branch_names):
                next_branch = None
            else:
                next_branch = sorted_branch_names[to_index]

        branches = {name: branch.to_dict() if branch else None
                    for (name, branch) in branches.items()}

        return {
                'id': snapshot_id,
                'branches': branches,
                'next_branch': next_branch,
                }

    def snapshot_get_random(self):
        """Finds a random snapshot id.

        Returns:
            a sha1_git
        """
        return random.choice(list(self._snapshots))

    def object_find_by_sha1_git(self, ids, db=None, cur=None):
        """Return the objects found with the given ids.

        Args:
            ids: a generator of sha1_gits

        Returns:
            dict: a mapping from id to the list of objects found. Each object
            found is itself a dict with keys:

            - sha1_git: the input id
            - type: the type of object found
            - id: the id of the object found
            - object_id: the numeric id of the object found.

        """
        ret = {}
        for id_ in ids:
            objs = self._objects.get(id_, [])
            ret[id_] = [{
                    'sha1_git': id_,
                    'type': obj[0],
                    'id': obj[1],
                    'object_id': id_,
                    } for obj in objs]
        return ret

    def _convert_origin(self, t):
        if t is None:
            return None

        return t.to_dict()

    def origin_get(self, origins):
        """Return origins, either all identified by their ids or all
        identified by urls.

        Args:
            origin: a list of dictionaries representing the individual
                origins to find.
                These dicts have either the key url (and optionally type):

                - url (bytes): the url the origin points to

                or the id:

                - id (int): the origin's identifier

        Returns:
            dict: the origin dictionary with the keys:

            - id: origin's id
            - url: origin's url

        Raises:
            ValueError: if the keys does not match (url and type) nor id.

        """
        if isinstance(origins, dict):
            # Old API
            return_single = True
            origins = [origins]
        else:
            return_single = False

        # Sanity check to be error-compatible with the pgsql backend
        if any('id' in origin for origin in origins) \
                and not all('id' in origin for origin in origins):
            raise ValueError(
                'Either all origins or none at all should have an "id".')
        if any('url' in origin for origin in origins) \
                and not all('url' in origin for origin in origins):
            raise ValueError(
                'Either all origins or none at all should have '
                'an "url" key.')

        results = []
        for origin in origins:
            result = None
            if 'url' in origin:
                if origin['url'] in self._origins:
                    result = self._origins[origin['url']]
            else:
                raise ValueError(
                    'Origin must have an url.')
            results.append(self._convert_origin(result))

        if return_single:
            assert len(results) == 1
            return results[0]
        else:
            return results

    def origin_get_by_sha1(self, sha1s):
        """Return origins, identified by the sha1 of their URLs.

        Args:
            sha1s (list[bytes]): a list of sha1s

        Yields:
            dicts containing origin information as returned
            by :meth:`swh.storage.in_memory.Storage.origin_get`, or None if an
            origin matching the sha1 is not found.
        """
        return [
            self._convert_origin(self._origins_by_sha1.get(sha1))
            for sha1 in sha1s
        ]

    def origin_get_range(self, origin_from=1, origin_count=100):
        """Retrieve ``origin_count`` origins whose ids are greater
        or equal than ``origin_from``.

        Origins are sorted by id before retrieving them.

        Args:
            origin_from (int): the minimum id of origins to retrieve
            origin_count (int): the maximum number of origins to retrieve

        Yields:
            dicts containing origin information as returned
            by :meth:`swh.storage.in_memory.Storage.origin_get`, plus
            an 'id' key.
        """
        origin_from = max(origin_from, 1)
        if origin_from <= len(self._origins_by_id):
            max_idx = origin_from + origin_count - 1
            if max_idx > len(self._origins_by_id):
                max_idx = len(self._origins_by_id)
            for idx in range(origin_from-1, max_idx):
                origin = self._convert_origin(
                    self._origins[self._origins_by_id[idx]])
                yield {'id': idx+1, **origin}

    def origin_list(self, page_token: Optional[str] = None, limit: int = 100
                    ) -> dict:
        """Returns the list of origins

        Args:
            page_token: opaque token used for pagination.
            limit: the maximum number of results to return

        Returns:
            dict: dict with the following keys:
              - **next_page_token** (str, optional): opaque token to be used as
                `page_token` for retrieving the next page. if absent, there is
                no more pages to gather.
              - **origins** (List[dict]): list of origins, as returned by
                `origin_get`.
        """
        origin_urls = sorted(self._origins)
        if page_token:
            from_ = bisect.bisect_left(origin_urls, page_token)
        else:
            from_ = 0

        result = {
            'origins': [{'url': origin_url}
                        for origin_url in origin_urls[from_:from_+limit]]
        }

        if from_+limit < len(origin_urls):
            result['next_page_token'] = origin_urls[from_+limit]

        return result

    def origin_search(self, url_pattern, offset=0, limit=50,
                      regexp=False, with_visit=False, db=None, cur=None):
        """Search for origins whose urls contain a provided string pattern
        or match a provided regular expression.
        The search is performed in a case insensitive way.

        Args:
            url_pattern (str): the string pattern to search for in origin urls
            offset (int): number of found origins to skip before returning
                results
            limit (int): the maximum number of found origins to return
            regexp (bool): if True, consider the provided pattern as a regular
                expression and return origins whose urls match it
            with_visit (bool): if True, filter out origins with no visit

        Returns:
            An iterable of dict containing origin information as returned
            by :meth:`swh.storage.storage.Storage.origin_get`.
        """
        origins = map(self._convert_origin, self._origins.values())
        if regexp:
            pat = re.compile(url_pattern)
            origins = [orig for orig in origins if pat.search(orig['url'])]
        else:
            origins = [orig for orig in origins if url_pattern in orig['url']]
        if with_visit:
            origins = [
                orig for orig in origins
                if len(self._origin_visits[orig['url']]) > 0 and
                set(ov.snapshot
                    for ov in self._origin_visits[orig['url']]
                    if ov.snapshot) &
                set(self._snapshots)]

        return origins[offset:offset+limit]

    def origin_count(self, url_pattern, regexp=False, with_visit=False,
                     db=None, cur=None):
        """Count origins whose urls contain a provided string pattern
        or match a provided regular expression.
        The pattern search in origin urls is performed in a case insensitive
        way.

        Args:
            url_pattern (str): the string pattern to search for in origin urls
            regexp (bool): if True, consider the provided pattern as a regular
                expression and return origins whose urls match it
            with_visit (bool): if True, filter out origins with no visit

        Returns:
            int: The number of origins matching the search criterion.
        """
        return len(self.origin_search(url_pattern, regexp=regexp,
                                      with_visit=with_visit,
                                      limit=len(self._origins)))

    def origin_add(self, origins):
        """Add origins to the storage

        Args:
            origins: list of dictionaries representing the individual origins,
                with the following keys:

                - url (bytes): the url the origin points to

        Returns:
            list: given origins as dict updated with their id

        """
        origins = copy.deepcopy(origins)
        for origin in origins:
            self.origin_add_one(origin)
        return origins

    def origin_add_one(self, origin):
        """Add origin to the storage

        Args:
            origin: dictionary representing the individual origin to add. This
                dict has the following keys:

                - url (bytes): the url the origin points to

        Returns:
            the id of the added origin, or of the identical one that already
            exists.

        """
        origin = Origin.from_dict(origin)
        if origin.url not in self._origins:
            if self.journal_writer:
                self.journal_writer.write_addition('origin', origin)

            # generate an origin_id because it is needed by origin_get_range.
            # TODO: remove this when we remove origin_get_range
            origin_id = len(self._origins) + 1
            self._origins_by_id.append(origin.url)
            assert len(self._origins_by_id) == origin_id

            self._origins[origin.url] = origin
            self._origins_by_sha1[origin_url_to_sha1(origin.url)] = origin
            self._origin_visits[origin.url] = []
            self._objects[origin.url].append(('origin', origin.url))

        return origin.url

    def origin_visit_add(self, origin, date, type):
        """Add an origin_visit for the origin at date with status 'ongoing'.

        Args:
            origin (str): visited origin's identifier or URL
            date (Union[str,datetime]): timestamp of such visit
            type (str): the type of loader used for the visit (hg, git, ...)

        Returns:
            dict: dictionary with keys origin and visit where:

            - origin: origin's identifier
            - visit: the visit's identifier for the new visit occurrence

        """
        origin_url = origin
        if origin_url is None:
            raise ValueError('Unknown origin.')

        if isinstance(date, str):
            # FIXME: Converge on iso8601 at some point
            date = dateutil.parser.parse(date)
        elif not isinstance(date, datetime.datetime):
            raise TypeError('date must be a datetime or a string.')

        visit_ret = None
        if origin_url in self._origins:
            origin = self._origins[origin_url]
            # visit ids are in the range [1, +inf[
            visit_id = len(self._origin_visits[origin_url]) + 1
            status = 'ongoing'
            visit = OriginVisit(
                origin=origin.url,
                date=date,
                type=type,
                status=status,
                snapshot=None,
                metadata=None,
                visit=visit_id,
            )
            self._origin_visits[origin_url].append(visit)
            visit_ret = {
                'origin': origin.url,
                'visit': visit_id,
            }

            self._objects[(origin_url, visit_id)].append(
                ('origin_visit', None))

            if self.journal_writer:
                self.journal_writer.write_addition('origin_visit', visit)

        return visit_ret

    def origin_visit_update(self, origin, visit_id, status=None,
                            metadata=None, snapshot=None):
        """Update an origin_visit's status.

        Args:
            origin (str): visited origin's URL
            visit_id (int): visit's identifier
            status: visit's new status
            metadata: data associated to the visit
            snapshot (sha1_git): identifier of the snapshot to add to
                the visit

        Returns:
            None

        """
        if not isinstance(origin, str):
            raise TypeError('origin must be a string, not %r' % (origin,))
        origin_url = self._get_origin_url(origin)
        if origin_url is None:
            raise ValueError('Unknown origin.')

        try:
            visit = self._origin_visits[origin_url][visit_id-1]
        except IndexError:
            raise ValueError('Unknown visit_id for this origin') \
                from None

        updates = {}
        if status:
            updates['status'] = status
        if metadata:
            updates['metadata'] = metadata
        if snapshot:
            updates['snapshot'] = snapshot

        visit = attr.evolve(visit, **updates)

        if self.journal_writer:
            self.journal_writer.write_update('origin_visit', visit)

        self._origin_visits[origin_url][visit_id-1] = visit

    def origin_visit_upsert(self, visits):
        """Add a origin_visits with a specific id and with all its data.
        If there is already an origin_visit with the same
        `(origin_url, visit_id)`, updates it instead of inserting a new one.

        Args:
            visits: iterable of dicts with keys:

                origin: origin url
                visit: origin visit id
                type: type of loader used for the visit
                date: timestamp of such visit
                status: Visit's new status
                metadata: Data associated to the visit
                snapshot (sha1_git): identifier of the snapshot to add to
                    the visit
        """
        for visit in visits:
            if not isinstance(visit['origin'], str):
                raise TypeError("visit['origin'] must be a string, not %r"
                                % (visit['origin'],))
        visits = [OriginVisit.from_dict(d) for d in visits]

        if self.journal_writer:
            for visit in visits:
                self.journal_writer.write_addition('origin_visit', visit)

        for visit in visits:
            visit_id = visit.visit
            origin_url = visit.origin

            visit = attr.evolve(visit, origin=origin_url)

            self._objects[(origin_url, visit_id)].append(
                ('origin_visit', None))

            while len(self._origin_visits[origin_url]) <= visit_id:
                self._origin_visits[origin_url].append(None)

            self._origin_visits[origin_url][visit_id-1] = visit

    def _convert_visit(self, visit):
        if visit is None:
            return

        visit = visit.to_dict()

        return visit

    def origin_visit_get(self, origin, last_visit=None, limit=None):
        """Retrieve all the origin's visit's information.

        Args:
            origin (int): the origin's identifier
            last_visit (int): visit's id from which listing the next ones,
                default to None
            limit (int): maximum number of results to return,
                default to None

        Yields:
            List of visits.

        """
        origin_url = self._get_origin_url(origin)
        if origin_url in self._origin_visits:
            visits = self._origin_visits[origin_url]
            if last_visit is not None:
                visits = visits[last_visit:]
            if limit is not None:
                visits = visits[:limit]
            for visit in visits:
                if not visit:
                    continue
                visit_id = visit.visit

                yield self._convert_visit(
                    self._origin_visits[origin_url][visit_id-1])

    def origin_visit_find_by_date(self, origin, visit_date):
        """Retrieves the origin visit whose date is closest to the provided
        timestamp.
        In case of a tie, the visit with largest id is selected.

        Args:
            origin (str): The occurrence's origin (URL).
            target (datetime): target timestamp

        Returns:
            A visit.

        """
        origin_url = self._get_origin_url(origin)
        if origin_url in self._origin_visits:
            visits = self._origin_visits[origin_url]
            visit = min(
                visits,
                key=lambda v: (abs(v.date - visit_date), -v.visit))
            return self._convert_visit(visit)

    def origin_visit_get_by(self, origin, visit):
        """Retrieve origin visit's information.

        Args:
            origin (int): the origin's identifier

        Returns:
            The information on that particular (origin, visit) or None if
            it does not exist

        """
        origin_url = self._get_origin_url(origin)
        if origin_url in self._origin_visits and \
           visit <= len(self._origin_visits[origin_url]):
            return self._convert_visit(
                self._origin_visits[origin_url][visit-1])

    def origin_visit_get_latest(
            self, origin, allowed_statuses=None, require_snapshot=False):
        """Get the latest origin visit for the given origin, optionally
        looking only for those with one of the given allowed_statuses
        or for those with a known snapshot.

        Args:
            origin (str): the origin's URL
            allowed_statuses (list of str): list of visit statuses considered
                to find the latest visit. For instance,
                ``allowed_statuses=['full']`` will only consider visits that
                have successfully run to completion.
            require_snapshot (bool): If True, only a visit with a snapshot
                will be returned.
        Returns:
            dict: a dict with the following keys:

                origin: the URL of the origin
                visit: origin visit id
                type: type of loader used for the visit
                date: timestamp of such visit
                status: Visit's new status
                metadata: Data associated to the visit
                snapshot (Optional[sha1_git]): identifier of the snapshot
                    associated to the visit
        """
        origin = self._origins.get(origin)
        if not origin:
            return
        visits = self._origin_visits[origin.url]
        if allowed_statuses is not None:
            visits = [visit for visit in visits
                      if visit.status in allowed_statuses]
        if require_snapshot:
            visits = [visit for visit in visits
                      if visit.snapshot]

        visit = max(
            visits, key=lambda v: (v.date, v.visit), default=None)
        return self._convert_visit(visit)

    def _select_random_origin_visit_by_type(self, type: str) -> str:
        """Select randomly an origin visit """
        while True:
            url = random.choice(list(self._origin_visits.keys()))
            random_origin_visits = self._origin_visits[url]
            if random_origin_visits[0].type == type:
                return url

    def origin_visit_get_random(self, type: str) -> Mapping[str, Any]:
        """Randomly select one successful origin visit with <type>
        made in the last 3 months.

        Returns:
            dict representing an origin visit, in the same format as
            `origin_visit_get`.

        """
        random_visit: Dict[str, Any] = {}
        if not self._origin_visits:  # empty dataset
            return random_visit
        url = self._select_random_origin_visit_by_type(type)
        random_origin_visits = copy.deepcopy(self._origin_visits[url])
        random_origin_visits.reverse()
        back_in_the_day = now() - timedelta(weeks=12)  # 3 months back
        # This should be enough for tests
        for visit in random_origin_visits:
            if visit.date > back_in_the_day and visit.status == 'full':
                random_visit = visit.to_dict()
                break
        return random_visit

    def stat_counters(self):
        """compute statistics about the number of tuples in various tables

        Returns:
            dict: a dictionary mapping textual labels (e.g., content) to
            integer values (e.g., the number of tuples in table content)

        """
        keys = (
            'content',
            'directory',
            'origin',
            'origin_visit',
            'person',
            'release',
            'revision',
            'skipped_content',
            'snapshot'
            )
        stats = {key: 0 for key in keys}
        stats.update(collections.Counter(
            obj_type
            for (obj_type, obj_id)
            in itertools.chain(*self._objects.values())))
        return stats

    def refresh_stat_counters(self):
        """Recomputes the statistics for `stat_counters`."""
        pass

    def origin_metadata_add(self, origin_url, ts, provider, tool, metadata,
                            db=None, cur=None):
        """ Add an origin_metadata for the origin at ts with provenance and
        metadata.

        Args:
            origin_url (str): the origin url for which the metadata is added
            ts (datetime): timestamp of the found metadata
            provider: id of the provider of metadata (ex:'hal')
            tool: id of the tool used to extract metadata
            metadata (jsonb): the metadata retrieved at the time and location
        """
        if not isinstance(origin_url, str):
            raise TypeError('origin_id must be str, not %r' % (origin_url,))

        if isinstance(ts, str):
            ts = dateutil.parser.parse(ts)

        origin_metadata = {
                'origin_url': origin_url,
                'discovery_date': ts,
                'tool_id': tool,
                'metadata': metadata,
                'provider_id': provider,
                }
        self._origin_metadata[origin_url].append(origin_metadata)
        return None

    def origin_metadata_get_by(self, origin_url, provider_type=None, db=None,
                               cur=None):
        """Retrieve list of all origin_metadata entries for the origin_url

        Args:
            origin_url (str): the origin's url
            provider_type (str): (optional) type of provider

        Returns:
            list of dicts: the origin_metadata dictionary with the keys:

            - origin_url (int): origin's URL
            - discovery_date (datetime): timestamp of discovery
            - tool_id (int): metadata's extracting tool
            - metadata (jsonb)
            - provider_id (int): metadata's provider
            - provider_name (str)
            - provider_type (str)
            - provider_url (str)

        """
        if not isinstance(origin_url, str):
            raise TypeError('origin_url must be str, not %r' % (origin_url,))
        metadata = []
        for item in self._origin_metadata[origin_url]:
            item = copy.deepcopy(item)
            provider = self.metadata_provider_get(item['provider_id'])
            for attr_name in ('name', 'type', 'url'):
                item['provider_' + attr_name] = \
                    provider['provider_' + attr_name]
            metadata.append(item)
        return metadata

    def tool_add(self, tools):
        """Add new tools to the storage.

        Args:
            tools (iterable of :class:`dict`): Tool information to add to
              storage. Each tool is a :class:`dict` with the following keys:

              - name (:class:`str`): name of the tool
              - version (:class:`str`): version of the tool
              - configuration (:class:`dict`): configuration of the tool,
                must be json-encodable

        Returns:
            :class:`dict`: All the tools inserted in storage
            (including the internal ``id``). The order of the list is not
            guaranteed to match the order of the initial list.

        """
        inserted = []
        for tool in tools:
            key = self._tool_key(tool)
            assert 'id' not in tool
            record = copy.deepcopy(tool)
            record['id'] = key  # TODO: remove this
            if key not in self._tools:
                self._tools[key] = record
            inserted.append(copy.deepcopy(self._tools[key]))

        return inserted

    def tool_get(self, tool):
        """Retrieve tool information.

        Args:
            tool (dict): Tool information we want to retrieve from storage.
              The dicts have the same keys as those used in :func:`tool_add`.

        Returns:
            dict: The full tool information if it exists (``id`` included),
            None otherwise.

        """
        return self._tools.get(self._tool_key(tool))

    def metadata_provider_add(self, provider_name, provider_type, provider_url,
                              metadata):
        """Add a metadata provider.

        Args:
            provider_name (str): Its name
            provider_type (str): Its type
            provider_url (str): Its URL
            metadata: JSON-encodable object

        Returns:
            an identifier of the provider
        """
        provider = {
                'provider_name': provider_name,
                'provider_type': provider_type,
                'provider_url': provider_url,
                'metadata': metadata,
                }
        key = self._metadata_provider_key(provider)
        provider['id'] = key
        self._metadata_providers[key] = provider
        return key

    def metadata_provider_get(self, provider_id, db=None, cur=None):
        """Get a metadata provider

        Args:
            provider_id: Its identifier, as given by `metadata_provider_add`.

        Returns:
            dict: same as `metadata_provider_add`;
                  or None if it does not exist.
        """
        return self._metadata_providers.get(provider_id)

    def metadata_provider_get_by(self, provider, db=None, cur=None):
        """Get a metadata provider

        Args:
            provider_name: Its name
            provider_url: Its URL

        Returns:
            dict: same as `metadata_provider_add`;
                  or None if it does not exist.
        """
        key = self._metadata_provider_key(provider)
        return self._metadata_providers.get(key)

    def _get_origin_url(self, origin):
        if isinstance(origin, str):
            return origin
        else:
            raise TypeError('origin must be a string.')

    def _person_add(self, person):
        """Add a person in storage.

        Note: Private method, do not use outside of this class.

        Args:
            person: dictionary with keys fullname, name and email.

        """
        key = ('person', person.fullname)
        if key not in self._objects:
            person_id = len(self._persons) + 1
            self._persons.append(person)
            self._objects[key].append(('person', person_id))
        else:
            person_id = self._objects[key][0][1]
            person = self._persons[person_id-1]
        return person

    @staticmethod
    def _content_key(content):
        """A stable key for a content"""
        return tuple(getattr(content, key)
                     for key in sorted(DEFAULT_ALGORITHMS))

    @staticmethod
    def _content_key_algorithm(content):
        """ A stable key and the algorithm for a content"""
        if isinstance(content, Content):
            content = content.to_dict()
        return tuple((content.get(key), key)
                     for key in sorted(DEFAULT_ALGORITHMS))

    @staticmethod
    def _tool_key(tool):
        return '%r %r %r' % (tool['name'], tool['version'],
                             tuple(sorted(tool['configuration'].items())))

    @staticmethod
    def _metadata_provider_key(provider):
        return '%r %r' % (provider['provider_name'], provider['provider_url'])



# Copyright (C) 2015-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging

import click

from swh.core.cli import CONTEXT_SETTINGS
from swh.storage.api.server import load_and_check_config, app


@click.group(name='storage', context_settings=CONTEXT_SETTINGS)
@click.pass_context
def storage(ctx):
    '''Software Heritage Storage tools.'''
    pass


@storage.command(name='rpc-serve')
@click.argument('config-path', required=True)
@click.option('--host', default='0.0.0.0',
              metavar='IP', show_default=True,
              help="Host ip address to bind the server on")
@click.option('--port', default=5002, type=click.INT,
              metavar='PORT', show_default=True,
              help="Binding port of the server")
@click.option('--debug/--no-debug', default=True,
              help="Indicates if the server should run in debug mode")
@click.pass_context
def serve(ctx, config_path, host, port, debug):
    '''Software Heritage Storage RPC server.

    Do NOT use this in a production environment.
    '''
    if 'log_level' in ctx.obj:
        logging.getLogger('werkzeug').setLevel(ctx.obj['log_level'])
    api_cfg = load_and_check_config(config_path, type='any')
    app.config.update(api_cfg)
    app.run(host, port=int(port), debug=bool(debug))


def main():
    logging.basicConfig()
    return serve(auto_envvar_prefix='SWH_STORAGE')


if __name__ == '__main__':
    main()



# Copyright (C) 2020 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import psycopg2
import pytest

from typing import Dict

from unittest.mock import call

from swh.storage import HashCollision
from swh.storage.retry import RetryingProxyStorage


@pytest.fixture
def swh_storage():
    return RetryingProxyStorage(storage={'cls': 'memory'})


def test_retrying_proxy_storage_content_add(swh_storage, sample_data):
    """Standard content_add works as before

    """
    sample_content = sample_data['content'][0]

    content = next(swh_storage.content_get([sample_content['sha1']]))
    assert not content

    s = swh_storage.content_add([sample_content])
    assert s == {
        'content:add': 1,
        'content:add:bytes': sample_content['length'],
        'skipped_content:add': 0
    }

    content = next(swh_storage.content_get([sample_content['sha1']]))
    assert content['sha1'] == sample_content['sha1']


def test_retrying_proxy_storage_content_add_with_retry(
        swh_storage, sample_data, mocker):
    """Multiple retries for hash collision and psycopg2 error but finally ok

    """
    mock_memory = mocker.patch('swh.storage.in_memory.Storage.content_add')
    mock_memory.side_effect = [
        # first try goes ko
        HashCollision('content hash collision'),
        # second try goes ko
        psycopg2.IntegrityError('content already inserted'),
        # ok then!
        {'content:add': 1}
    ]

    sample_content = sample_data['content'][0]

    content = next(swh_storage.content_get([sample_content['sha1']]))
    assert not content

    s = swh_storage.content_add([sample_content])
    assert s == {'content:add': 1}

    assert mock_memory.has_calls([
        call([sample_content]),
        call([sample_content]),
        call([sample_content]),
    ])


def test_retrying_proxy_swh_storage_content_add_failure(
        swh_storage, sample_data, mocker):
    """Unfiltered errors are raising without retry

    """
    mock_memory = mocker.patch('swh.storage.in_memory.Storage.content_add')
    mock_memory.side_effect = ValueError('Refuse to add content always!')

    sample_content = sample_data['content'][0]

    content = next(swh_storage.content_get([sample_content['sha1']]))
    assert not content

    with pytest.raises(ValueError, match='Refuse to add'):
        swh_storage.content_add([sample_content])

    assert mock_memory.call_count == 1


def test_retrying_proxy_storage_content_add_metadata(swh_storage, sample_data):
    """Standard content_add_metadata works as before

    """
    sample_content = sample_data['content_metadata'][0]

    pk = sample_content['sha1']
    content_metadata = swh_storage.content_get_metadata([pk])
    assert not content_metadata[pk]

    s = swh_storage.content_add_metadata([sample_content])
    assert s == {
        'content:add': 1,
        'skipped_content:add': 0
    }

    content_metadata = swh_storage.content_get_metadata([pk])
    assert len(content_metadata[pk]) == 1
    assert content_metadata[pk][0]['sha1'] == pk


def test_retrying_proxy_storage_content_add_metadata_with_retry(
        swh_storage, sample_data, mocker):
    """Multiple retries for hash collision and psycopg2 error but finally ok

    """
    mock_memory = mocker.patch(
        'swh.storage.in_memory.Storage.content_add_metadata')
    mock_memory.side_effect = [
        # first try goes ko
        HashCollision('content_metadata hash collision'),
        # second try goes ko
        psycopg2.IntegrityError('content_metadata already inserted'),
        # ok then!
        {'content:add': 1}
    ]

    sample_content = sample_data['content_metadata'][0]

    s = swh_storage.content_add_metadata([sample_content])
    assert s == {'content:add': 1}

    assert mock_memory.has_calls([
        call([sample_content]),
        call([sample_content]),
        call([sample_content]),
    ])


def test_retrying_proxy_swh_storage_content_add_metadata_failure(
        swh_storage, sample_data, mocker):
    """Unfiltered errors are raising without retry

    """
    mock_memory = mocker.patch(
        'swh.storage.in_memory.Storage.content_add_metadata')
    mock_memory.side_effect = ValueError('Refuse to add content_metadata!')

    sample_content = sample_data['content_metadata'][0]
    pk = sample_content['sha1']

    content_metadata = swh_storage.content_get_metadata([pk])
    assert not content_metadata[pk]

    with pytest.raises(ValueError, match='Refuse to add'):
        swh_storage.content_add_metadata([sample_content])

    assert mock_memory.call_count == 1


def test_retrying_proxy_swh_storage_origin_add_one(swh_storage, sample_data):
    """Standard origin_add_one works as before

    """
    sample_origin = sample_data['origin'][0]

    origin = swh_storage.origin_get(sample_origin)
    assert not origin

    r = swh_storage.origin_add_one(sample_origin)
    assert r == sample_origin['url']

    origin = swh_storage.origin_get(sample_origin)
    assert origin['url'] == sample_origin['url']


def test_retrying_proxy_swh_storage_origin_add_one_retry(
        swh_storage, sample_data, mocker):
    """Multiple retries for hash collision and psycopg2 error but finally ok

    """
    sample_origin = sample_data['origin'][1]
    mock_memory = mocker.patch('swh.storage.in_memory.Storage.origin_add_one')
    mock_memory.side_effect = [
        # first try goes ko
        HashCollision('origin hash collision'),
        # second try goes ko
        psycopg2.IntegrityError('origin already inserted'),
        # ok then!
        sample_origin['url']
    ]

    origin = swh_storage.origin_get(sample_origin)
    assert not origin

    r = swh_storage.origin_add_one(sample_origin)
    assert r == sample_origin['url']

    assert mock_memory.has_calls([
        call([sample_origin]),
        call([sample_origin]),
        call([sample_origin]),
    ])


def test_retrying_proxy_swh_storage_origin_add_one_failure(
        swh_storage, sample_data, mocker):
    """Unfiltered errors are raising without retry

    """
    mock_memory = mocker.patch('swh.storage.in_memory.Storage.origin_add_one')
    mock_memory.side_effect = ValueError('Refuse to add origin always!')

    sample_origin = sample_data['origin'][0]

    origin = swh_storage.origin_get(sample_origin)
    assert not origin

    with pytest.raises(ValueError, match='Refuse to add'):
        swh_storage.origin_add_one([sample_origin])

    assert mock_memory.call_count == 1


def test_retrying_proxy_swh_storage_origin_visit_add(swh_storage, sample_data):
    """Standard origin_visit_add works as before

    """
    sample_origin = sample_data['origin'][0]
    swh_storage.origin_add_one(sample_origin)
    origin_url = sample_origin['url']

    origin = list(swh_storage.origin_visit_get(origin_url))
    assert not origin

    origin_visit = swh_storage.origin_visit_add(
        origin_url, '2020-01-01', 'hg')
    assert origin_visit['origin'] == origin_url
    assert isinstance(origin_visit['visit'], int)

    origin_visit = next(swh_storage.origin_visit_get(origin_url))
    assert origin_visit['origin'] == origin_url
    assert isinstance(origin_visit['visit'], int)


def test_retrying_proxy_swh_storage_origin_visit_add_retry(
        swh_storage, sample_data, mocker):
    """Multiple retries for hash collision and psycopg2 error but finally ok

    """
    sample_origin = sample_data['origin'][1]
    swh_storage.origin_add_one(sample_origin)
    origin_url = sample_origin['url']

    mock_memory = mocker.patch(
        'swh.storage.in_memory.Storage.origin_visit_add')
    mock_memory.side_effect = [
        # first try goes ko
        HashCollision('origin hash collision'),
        # second try goes ko
        psycopg2.IntegrityError('origin already inserted'),
        # ok then!
        {'origin': origin_url, 'visit': 1}
    ]

    origin = list(swh_storage.origin_visit_get(origin_url))
    assert not origin

    r = swh_storage.origin_visit_add(sample_origin, '2020-01-01', 'git')
    assert r == {'origin': origin_url, 'visit': 1}

    assert mock_memory.has_calls([
        call(sample_origin, '2020-01-01', 'git'),
        call(sample_origin, '2020-01-01', 'git'),
        call(sample_origin, '2020-01-01', 'git')
    ])


def test_retrying_proxy_swh_storage_origin_visit_add_failure(
        swh_storage, sample_data, mocker):
    """Unfiltered errors are raising without retry

    """
    mock_memory = mocker.patch(
        'swh.storage.in_memory.Storage.origin_visit_add')
    mock_memory.side_effect = ValueError('Refuse to add origin always!')

    origin_url = sample_data['origin'][0]['url']

    origin = list(swh_storage.origin_visit_get(origin_url))
    assert not origin

    with pytest.raises(ValueError, match='Refuse to add'):
        swh_storage.origin_visit_add(origin_url, '2020-01-31', 'svn')

    assert mock_memory.has_calls([
        call(origin_url, '2020-01-31', 'svn'),
    ])


def test_retrying_proxy_storage_tool_add(swh_storage, sample_data):
    """Standard tool_add works as before

    """
    sample_tool = sample_data['tool'][0]

    tool = swh_storage.tool_get(sample_tool)
    assert not tool

    tools = swh_storage.tool_add([sample_tool])
    assert tools
    tool = tools[0]
    tool.pop('id')
    assert tool == sample_tool

    tool = swh_storage.tool_get(sample_tool)
    tool.pop('id')
    assert tool == sample_tool


def test_retrying_proxy_storage_tool_add_with_retry(
        swh_storage, sample_data, mocker):
    """Multiple retries for hash collision and psycopg2 error but finally ok

    """
    sample_tool = sample_data['tool'][0]
    mock_memory = mocker.patch('swh.storage.in_memory.Storage.tool_add')
    mock_memory.side_effect = [
        # first try goes ko
        HashCollision('tool hash collision'),
        # second try goes ko
        psycopg2.IntegrityError('tool already inserted'),
        # ok then!
        [sample_tool]
    ]

    tool = swh_storage.tool_get(sample_tool)
    assert not tool

    tools = swh_storage.tool_add([sample_tool])
    assert tools == [sample_tool]

    assert mock_memory.has_calls([
        call([sample_tool]),
        call([sample_tool]),
        call([sample_tool]),
    ])


def test_retrying_proxy_swh_storage_tool_add_failure(
        swh_storage, sample_data, mocker):
    """Unfiltered errors are raising without retry

    """
    mock_memory = mocker.patch('swh.storage.in_memory.Storage.tool_add')
    mock_memory.side_effect = ValueError('Refuse to add tool always!')

    sample_tool = sample_data['tool'][0]

    tool = swh_storage.tool_get(sample_tool)
    assert not tool

    with pytest.raises(ValueError, match='Refuse to add'):
        swh_storage.tool_add([sample_tool])

    assert mock_memory.call_count == 1


def to_provider(provider: Dict) -> Dict:
    return {
        'provider_name': provider['name'],
        'provider_url': provider['url'],
        'provider_type': provider['type'],
        'metadata': provider['metadata'],
    }


def test_retrying_proxy_storage_metadata_provider_add(
        swh_storage, sample_data):
    """Standard metadata_provider_add works as before

    """
    provider = sample_data['provider'][0]
    provider_get = to_provider(provider)

    provider = swh_storage.metadata_provider_get_by(provider_get)
    assert not provider

    provider_id = swh_storage.metadata_provider_add(**provider_get)
    assert provider_id

    actual_provider = swh_storage.metadata_provider_get(provider_id)
    assert actual_provider
    actual_provider_id = actual_provider.pop('id')
    assert actual_provider_id == provider_id
    assert actual_provider == provider_get


def test_retrying_proxy_storage_metadata_provider_add_with_retry(
        swh_storage, sample_data, mocker):
    """Multiple retries for hash collision and psycopg2 error but finally ok

    """
    provider = sample_data['provider'][0]
    provider_get = to_provider(provider)

    mock_memory = mocker.patch(
        'swh.storage.in_memory.Storage.metadata_provider_add')
    mock_memory.side_effect = [
        # first try goes ko
        HashCollision('provider_id hash collision'),
        # second try goes ko
        psycopg2.IntegrityError('provider_id already inserted'),
        # ok then!
        'provider_id',
    ]

    provider = swh_storage.metadata_provider_get_by(provider_get)
    assert not provider

    provider_id = swh_storage.metadata_provider_add(**provider_get)
    assert provider_id == 'provider_id'

    assert mock_memory.has_calls([
        call(**provider_get),
        call(**provider_get),
        call(**provider_get),
    ])


def test_retrying_proxy_swh_storage_metadata_provider_add_failure(
        swh_storage, sample_data, mocker):
    """Unfiltered errors are raising without retry

    """
    mock_memory = mocker.patch(
        'swh.storage.in_memory.Storage.metadata_provider_add')
    mock_memory.side_effect = ValueError('Refuse to add provider_id always!')

    provider = sample_data['provider'][0]
    provider_get = to_provider(provider)

    provider_id = swh_storage.metadata_provider_get_by(provider_get)
    assert not provider_id

    with pytest.raises(ValueError, match='Refuse to add'):
        swh_storage.metadata_provider_add(**provider_get)

    assert mock_memory.call_count == 1


def test_retrying_proxy_storage_origin_metadata_add(
        swh_storage, sample_data):
    """Standard origin_metadata_add works as before

    """
    ori_meta = sample_data['origin_metadata'][0]
    origin = ori_meta['origin']
    swh_storage.origin_add_one(origin)
    provider_get = to_provider(ori_meta['provider'])
    provider_id = swh_storage.metadata_provider_add(**provider_get)

    origin_metadata = swh_storage.origin_metadata_get_by(origin['url'])
    assert not origin_metadata

    swh_storage.origin_metadata_add(
        origin['url'], ori_meta['discovery_date'],
        provider_id, ori_meta['tool'], ori_meta['metadata'])

    origin_metadata = swh_storage.origin_metadata_get_by(
        origin['url'])
    assert origin_metadata


def test_retrying_proxy_storage_origin_metadata_add_with_retry(
        swh_storage, sample_data, mocker):
    """Multiple retries for hash collision and psycopg2 error but finally ok

    """
    ori_meta = sample_data['origin_metadata'][0]
    origin = ori_meta['origin']
    swh_storage.origin_add_one(origin)
    provider_get = to_provider(ori_meta['provider'])
    provider_id = swh_storage.metadata_provider_add(**provider_get)
    mock_memory = mocker.patch(
        'swh.storage.in_memory.Storage.origin_metadata_add')

    mock_memory.side_effect = [
        # first try goes ko
        HashCollision('provider_id hash collision'),
        # second try goes ko
        psycopg2.IntegrityError('provider_id already inserted'),
        # ok then!
        None
    ]

    url = origin['url']
    ts = ori_meta['discovery_date']
    tool_id = ori_meta['tool']
    metadata = ori_meta['metadata']

    # No exception raised as insertion finally came through
    swh_storage.origin_metadata_add(url, ts, provider_id, tool_id, metadata)

    mock_memory.assert_has_calls([  # 3 calls, as long as error raised
        call(url, ts, provider_id, tool_id, metadata),
        call(url, ts, provider_id, tool_id, metadata),
        call(url, ts, provider_id, tool_id, metadata)
    ])


def test_retrying_proxy_swh_storage_origin_metadata_add_failure(
        swh_storage, sample_data, mocker):
    """Unfiltered errors are raising without retry

    """
    mock_memory = mocker.patch(
        'swh.storage.in_memory.Storage.origin_metadata_add')
    mock_memory.side_effect = ValueError('Refuse to add always!')

    ori_meta = sample_data['origin_metadata'][0]
    origin = ori_meta['origin']
    swh_storage.origin_add_one(origin)

    url = origin['url']
    ts = ori_meta['discovery_date']
    provider_id = 'provider_id'
    tool_id = ori_meta['tool']
    metadata = ori_meta['metadata']

    with pytest.raises(ValueError, match='Refuse to add'):
        swh_storage.origin_metadata_add(url, ts, provider_id, tool_id,
                                        metadata)

    assert mock_memory.call_count == 1


def test_retrying_proxy_swh_storage_origin_visit_update(
        swh_storage, sample_data):
    """Standard origin_visit_update works as before

    """
    sample_origin = sample_data['origin'][0]
    swh_storage.origin_add_one(sample_origin)
    origin_url = sample_origin['url']
    origin_visit = swh_storage.origin_visit_add(
        origin_url, '2020-01-01', 'hg')

    ov = next(swh_storage.origin_visit_get(origin_url))
    assert ov['origin'] == origin_url
    assert ov['visit'] == origin_visit['visit']
    assert ov['status'] == 'ongoing'
    assert ov['snapshot'] is None
    assert ov['metadata'] is None

    swh_storage.origin_visit_update(origin_url, ov['visit'], status='full')

    ov = next(swh_storage.origin_visit_get(origin_url))
    assert ov['origin'] == origin_url
    assert ov['visit'] == origin_visit['visit']
    assert ov['status'] == 'full'
    assert ov['snapshot'] is None
    assert ov['metadata'] is None


def test_retrying_proxy_swh_storage_origin_visit_update_retry(
        swh_storage, sample_data, mocker):
    """Multiple retries for hash collision and psycopg2 error but finally ok

    """
    sample_origin = sample_data['origin'][1]
    origin_url = sample_origin['url']

    mock_memory = mocker.patch(
        'swh.storage.in_memory.Storage.origin_visit_update')
    mock_memory.side_effect = [
        # first try goes ko
        HashCollision('origin hash collision'),
        # second try goes ko
        psycopg2.IntegrityError('origin already inserted'),
        # ok then!
        {'origin': origin_url, 'visit': 1}
    ]

    visit_id = 1
    swh_storage.origin_visit_update(origin_url, visit_id, status='full')

    assert mock_memory.has_calls([
        call(origin_url, visit_id, status='full'),
        call(origin_url, visit_id, status='full'),
        call(origin_url, visit_id, status='full'),
    ])


def test_retrying_proxy_swh_storage_origin_visit_update_failure(
        swh_storage, sample_data, mocker):
    """Unfiltered errors are raising without retry

    """
    mock_memory = mocker.patch(
        'swh.storage.in_memory.Storage.origin_visit_update')
    mock_memory.side_effect = ValueError('Refuse to add origin always!')
    origin_url = sample_data['origin'][0]['url']
    visit_id = 9

    with pytest.raises(ValueError, match='Refuse to add'):
        swh_storage.origin_visit_update(origin_url, visit_id, 'partial')

    assert mock_memory.call_count == 1


def test_retrying_proxy_storage_directory_add(swh_storage, sample_data):
    """Standard directory_add works as before

    """
    sample_dir = sample_data['directory'][0]

    directory = swh_storage.directory_get_random()  # no directory
    assert not directory

    s = swh_storage.directory_add([sample_dir])
    assert s == {
        'directory:add': 1,
    }

    directory_id = swh_storage.directory_get_random()  # only 1
    assert directory_id == sample_dir['id']


def test_retrying_proxy_storage_directory_add_with_retry(
        swh_storage, sample_data, mocker):
    """Multiple retries for hash collision and psycopg2 error but finally ok

    """
    mock_memory = mocker.patch('swh.storage.in_memory.Storage.directory_add')
    mock_memory.side_effect = [
        # first try goes ko
        HashCollision('directory hash collision'),
        # second try goes ko
        psycopg2.IntegrityError('directory already inserted'),
        # ok then!
        {'directory:add': 1}
    ]

    sample_dir = sample_data['directory'][1]

    directory_id = swh_storage.directory_get_random()  # no directory
    assert not directory_id

    s = swh_storage.directory_add([sample_dir])
    assert s == {
        'directory:add': 1,
    }

    assert mock_memory.has_calls([
        call([sample_dir]),
        call([sample_dir]),
        call([sample_dir]),
    ])


def test_retrying_proxy_swh_storage_directory_add_failure(
        swh_storage, sample_data, mocker):
    """Unfiltered errors are raising without retry

    """
    mock_memory = mocker.patch('swh.storage.in_memory.Storage.directory_add')
    mock_memory.side_effect = ValueError('Refuse to add directory always!')

    sample_dir = sample_data['directory'][0]

    directory_id = swh_storage.directory_get_random()  # no directory
    assert not directory_id

    with pytest.raises(ValueError, match='Refuse to add'):
        swh_storage.directory_add([sample_dir])

    assert mock_memory.call_count == 1


def test_retrying_proxy_storage_revision_add(swh_storage, sample_data):
    """Standard revision_add works as before

    """
    sample_rev = sample_data['revision'][0]

    revision = next(swh_storage.revision_get([sample_rev['id']]))
    assert not revision

    s = swh_storage.revision_add([sample_rev])
    assert s == {
        'revision:add': 1,
    }

    revision = next(swh_storage.revision_get([sample_rev['id']]))
    assert revision['id'] == sample_rev['id']


def test_retrying_proxy_storage_revision_add_with_retry(
        swh_storage, sample_data, mocker):
    """Multiple retries for hash collision and psycopg2 error but finally ok

    """
    mock_memory = mocker.patch('swh.storage.in_memory.Storage.revision_add')
    mock_memory.side_effect = [
        # first try goes ko
        HashCollision('revision hash collision'),
        # second try goes ko
        psycopg2.IntegrityError('revision already inserted'),
        # ok then!
        {'revision:add': 1}
    ]

    sample_rev = sample_data['revision'][0]

    revision = next(swh_storage.revision_get([sample_rev['id']]))
    assert not revision

    s = swh_storage.revision_add([sample_rev])
    assert s == {
        'revision:add': 1,
    }

    assert mock_memory.has_calls([
        call([sample_rev]),
        call([sample_rev]),
        call([sample_rev]),
    ])


def test_retrying_proxy_swh_storage_revision_add_failure(
        swh_storage, sample_data, mocker):
    """Unfiltered errors are raising without retry

    """
    mock_memory = mocker.patch('swh.storage.in_memory.Storage.revision_add')
    mock_memory.side_effect = ValueError('Refuse to add revision always!')

    sample_rev = sample_data['revision'][0]

    revision = next(swh_storage.revision_get([sample_rev['id']]))
    assert not revision

    with pytest.raises(ValueError, match='Refuse to add'):
        swh_storage.revision_add([sample_rev])

    assert mock_memory.call_count == 1


def test_retrying_proxy_storage_release_add(swh_storage, sample_data):
    """Standard release_add works as before

    """
    sample_rel = sample_data['release'][0]

    release = next(swh_storage.release_get([sample_rel['id']]))
    assert not release

    s = swh_storage.release_add([sample_rel])
    assert s == {
        'release:add': 1,
    }

    release = next(swh_storage.release_get([sample_rel['id']]))
    assert release['id'] == sample_rel['id']


def test_retrying_proxy_storage_release_add_with_retry(
        swh_storage, sample_data, mocker):
    """Multiple retries for hash collision and psycopg2 error but finally ok

    """
    mock_memory = mocker.patch('swh.storage.in_memory.Storage.release_add')
    mock_memory.side_effect = [
        # first try goes ko
        HashCollision('release hash collision'),
        # second try goes ko
        psycopg2.IntegrityError('release already inserted'),
        # ok then!
        {'release:add': 1}
    ]

    sample_rel = sample_data['release'][0]

    release = next(swh_storage.release_get([sample_rel['id']]))
    assert not release

    s = swh_storage.release_add([sample_rel])
    assert s == {
        'release:add': 1,
    }

    assert mock_memory.has_calls([
        call([sample_rel]),
        call([sample_rel]),
        call([sample_rel]),
    ])


def test_retrying_proxy_swh_storage_release_add_failure(
        swh_storage, sample_data, mocker):
    """Unfiltered errors are raising without retry

    """
    mock_memory = mocker.patch('swh.storage.in_memory.Storage.release_add')
    mock_memory.side_effect = ValueError('Refuse to add release always!')

    sample_rel = sample_data['release'][0]

    release = next(swh_storage.release_get([sample_rel['id']]))
    assert not release

    with pytest.raises(ValueError, match='Refuse to add'):
        swh_storage.release_add([sample_rel])

    assert mock_memory.call_count == 1


def test_retrying_proxy_storage_snapshot_add(swh_storage, sample_data):
    """Standard snapshot_add works as before

    """
    sample_snap = sample_data['snapshot'][0]

    snapshot = swh_storage.snapshot_get(sample_snap['id'])
    assert not snapshot

    s = swh_storage.snapshot_add([sample_snap])
    assert s == {
        'snapshot:add': 1,
    }

    snapshot = swh_storage.snapshot_get(sample_snap['id'])
    assert snapshot['id'] == sample_snap['id']


def test_retrying_proxy_storage_snapshot_add_with_retry(
        swh_storage, sample_data, mocker):
    """Multiple retries for hash collision and psycopg2 error but finally ok

    """
    mock_memory = mocker.patch('swh.storage.in_memory.Storage.snapshot_add')
    mock_memory.side_effect = [
        # first try goes ko
        HashCollision('snapshot hash collision'),
        # second try goes ko
        psycopg2.IntegrityError('snapshot already inserted'),
        # ok then!
        {'snapshot:add': 1}
    ]

    sample_snap = sample_data['snapshot'][0]

    snapshot = swh_storage.snapshot_get(sample_snap['id'])
    assert not snapshot

    s = swh_storage.snapshot_add([sample_snap])
    assert s == {
        'snapshot:add': 1,
    }

    assert mock_memory.has_calls([
        call([sample_snap]),
        call([sample_snap]),
        call([sample_snap]),
    ])


def test_retrying_proxy_swh_storage_snapshot_add_failure(
        swh_storage, sample_data, mocker):
    """Unfiltered errors are raising without retry

    """
    mock_memory = mocker.patch('swh.storage.in_memory.Storage.snapshot_add')
    mock_memory.side_effect = ValueError('Refuse to add snapshot always!')

    sample_snap = sample_data['snapshot'][0]

    snapshot = swh_storage.snapshot_get(sample_snap['id'])
    assert not snapshot

    with pytest.raises(ValueError, match='Refuse to add'):
        swh_storage.snapshot_add([sample_snap])

    assert mock_memory.call_count == 1



# Copyright (C) 2015-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import warnings

from . import storage

Storage = storage.Storage


class HashCollision(Exception):
    pass


STORAGE_IMPLEMENTATION = {
    'pipeline', 'local', 'remote', 'memory', 'filter', 'buffer', 'retry',
}


def get_storage(cls, **kwargs):
    """Get a storage object of class `storage_class` with arguments
    `storage_args`.

    Args:
        storage (dict): dictionary with keys:
        - cls (str): storage's class, either local, remote, memory, filter,
            buffer
        - args (dict): dictionary with keys

    Returns:
        an instance of swh.storage.Storage or compatible class

    Raises:
        ValueError if passed an unknown storage class.

    """
    if cls not in STORAGE_IMPLEMENTATION:
        raise ValueError('Unknown storage class `%s`. Supported: %s' % (
            cls, ', '.join(STORAGE_IMPLEMENTATION)))

    if 'args' in kwargs:
        warnings.warn(
            'Explicit "args" key is deprecated, use keys directly instead.',
            DeprecationWarning)
        kwargs = kwargs['args']

    if cls == 'pipeline':
        return get_storage_pipeline(**kwargs)

    if cls == 'remote':
        from .api.client import RemoteStorage as Storage
    elif cls == 'local':
        from .storage import Storage
    elif cls == 'memory':
        from .in_memory import Storage
    elif cls == 'filter':
        from .filter import FilteringProxyStorage as Storage
    elif cls == 'buffer':
        from .buffer import BufferingProxyStorage as Storage
    elif cls == 'retry':
        from .retry import RetryingProxyStorage as Storage

    return Storage(**kwargs)


def get_storage_pipeline(steps):
    """Recursively get a storage object that may use other storage objects
    as backends.

    Args:
        steps (List[dict]): List of dicts that may be used as kwargs for
            `get_storage`.

    Returns:
        an instance of swh.storage.Storage or compatible class

    Raises:
        ValueError if passed an unknown storage class.
    """
    storage_config = None
    for step in reversed(steps):
        if 'args' in step:
            warnings.warn(
                'Explicit "args" key is deprecated, use keys directly '
                'instead.',
                DeprecationWarning)
            step = {
                'cls': step['cls'],
                **step['args'],
            }
        if storage_config:
            step['storage'] = storage_config
        storage_config = step

    return get_storage(**storage_config)
